from plotext._default import default_monitor_class
from plotext._matrix import matrix_class
from plotext._build import build_class
import plotext._utility as ut
from copy import deepcopy
import math

# This file defines the monitor class, i.e. the plot, where actual data is plotted; the plot is build separately in the build class for clarity; here only the main tools and drawing methods are written

class monitor_class(build_class):
    
    def __init__(self):
        self.default = default_monitor_class() # default values
        self.labels_init()
        self.axes_init()
        self.color_init()
        self.data_init()
        self.matrix = matrix_class()

    def copy(self): # to deep copy 
        return deepcopy(self)

    def set_size(self, size): # called externally by the figure containing it, to pass the size
        self.size = size

    def set_date(self, date): # called externally by the figure containing it, to pass the date tools so that they share the same settings
        self.date = date

##############################################
#########    Internal Variables    ###########
##############################################

    def labels_init(self):
        self.title = None
        self.xlabel = [None, None]
        self.ylabel = [None, None]

    def axes_init(self):
        self.xscale = [self.default.xscale[0]] * 2 # the scale on x axis
        self.yscale = [self.default.xscale[0]] * 2

        self.xticks = self.default.xticks # xticks coordinates for both axes
        self.xlabels = self.default.xticks[:] # xlabels for both axes
        self.xfrequency = self.default.xfrequency # lower and upper xaxes ticks frequency
        self.xdirection = self.default.xdirection
        
        self.yticks = self.default.yticks
        self.ylabels = self.default.yticks[:]
        self.yfrequency = self.default.yfrequency # left and right yaxes ticks frequency
        self.ydirection = self.default.ydirection

        self.xaxes = self.default.xaxes # whatever to show the lower and upper x axis
        self.yaxes = self.default.yaxes # whatever to show the left and right y axis

        self.grid = self.default.grid # whatever to show the horizontal and vertical grid lines

    def color_init(self):
        self.set_theme('default')

    def data_init(self):
        self.xlim = [[None, None], [None, None]] # the x axis plot limits for lower and upper xside
        self.ylim = [[None, None], [None, None]] # the y axis plot limits for left and right yside

        self.fast_plot = False
        self.lines_init()
        self.text_init()
        self.draw_init()

    def lines_init(self):
        self.vcoord = [[], []] # those are user defined extra grid lines, vertical or horizontal, for each axis
        self.hcoord = [[], []]
        self.vcolors = [[], []] # their color
        self.hcolors = [[], []]

    def text_init(self):
        self.text = []
        self.tx = []
        self.ty = []
        self.txside = []
        self.tyside = []
        self.torien = []
        self.talign = []
        self.tfull = []
        self.tback = []
        self.tstyle = []

    def draw_init(self):  # Variables Set with Draw internal Arguments
        self.xside = [] # which side the x axis should go, for each plot (lower or upper)
        self.yside = [] # which side the y axis should go, for each plot (left or right)
        
        self.x = [] # list of x coordinates 
        self.y = [] # list of y coordinates
        self.x_date = [False, False] # True if x axis is for date time plots
        self.y_date = [False, False]
        self.signals = 0 # number of signals to plot

        self.lines = [] # whatever to draw lines between points

        self.marker = [] # list of markers used for each plot
        self.color = [] # list of marker colors used for each plot
        self.past_colors = []
        self.style = []

        self.fillx = [] # fill data vertically (till x axis)
        self.filly = [] # fill data horizontally (till y axis)

        self.label = [] # subplot list of labels

##############################################
#######    External Set Functions    #########
##############################################

    def set_title(self, title = None):
        self.title = self.set_label(title)

    def set_xlabel(self, label = None, xside = None):
        pos = self.xside_to_pos(xside)
        self.xlabel[pos] = self.set_label(label)
        
    def set_ylabel(self, label = None, yside = None):
        pos = self.yside_to_pos(yside)
        self.ylabel[pos] = self.set_label(label)
        
    def set_xlim(self, left = None, right = None, xside = None):
        left = self.date.string_to_time(left) if isinstance(left, str) else left
        right = self.date.string_to_time(right) if isinstance(right, str) else right
        left = None if left is None else float(left)
        right = None if right is None else float(right)
        xlim = [left, right]
        xlim = xlim if None in xlim else [min(xlim), max(xlim)]
        pos = self.xside_to_pos(xside)
        self.xlim[pos] = xlim

    def set_ylim(self, lower = None, upper = None, yside = None):
        lower = self.date.string_to_time(lower) if isinstance(lower, str) else lower
        upper = self.date.string_to_time(upper) if isinstance(upper, str) else upper
        lower = None if lower is None else float(lower)
        upper = None if upper is None else float(upper)
        ylim = [lower, upper]
        ylim = ylim if None in ylim else [min(ylim), max(ylim)]
        pos = self.yside_to_pos(yside)
        self.ylim[pos] = ylim

    def set_xscale(self, scale = None, xside = None):
        default_case = (scale is None or scale not in self.default.xscale)
        scale = self.default.xscale[0] if default_case else scale
        pos = self.xside_to_pos(xside)
        self.xscale[pos] = scale

    def set_yscale(self, scale = None, yside = None):
        default_case = (scale is None or scale not in self.default.yscale)
        scale = self.default.yscale[0] if default_case else scale
        pos = self.yside_to_pos(yside)
        self.yscale[pos] = scale

    def set_xticks(self, ticks = None, labels = None, xside = None):
        pos = self.xside_to_pos(xside)
        ticks = self.default.xticks[pos] if ticks is None else list(ticks)
        string_ticks = any([isinstance(el, str) for el in ticks])
        labels = ticks if string_ticks and labels is None else labels
        ticks = self.date.strings_to_time(ticks) if string_ticks else ticks
        labels = ut.get_labels(ticks) if labels is None else list(labels)
        labels = list(map(str, labels))
        ticks, labels = ut.brush(ticks, labels)
        self.xticks[pos] = ticks
        self.xlabels[pos] = labels
        self.xfrequency[pos] = self.xfrequency[pos] if ticks is None else len(ticks)

    def set_yticks(self, ticks = None, labels = None, yside = None):
        pos = self.yside_to_pos(yside)
        ticks = self.default.yticks[pos] if ticks is None else list(ticks)
        string_ticks = any([isinstance(el, str) for el in ticks])
        labels = ticks if string_ticks and labels is None else labels
        ticks = self.date.strings_to_time(ticks) if string_ticks else ticks
        labels = ut.get_labels(ticks) if labels is None else list(labels)
        labels = list(map(str, labels))
        ticks, labels = ut.brush(ticks, labels)
        self.yticks[pos] = ticks
        self.ylabels[pos] = labels
        self.yfrequency[pos] = self.yfrequency[pos] if ticks is None else len(ticks)

    def set_xfrequency(self, frequency = None, xside = None):
        pos = self.xside_to_pos(xside)
        frequency = self.default.xfrequency[pos] if frequency is None else int(frequency)
        self.xfrequency[pos] = frequency
        
    def set_yfrequency(self, frequency = None, yside = None):
        pos = self.yside_to_pos(yside)
        frequency = self.default.yfrequency[pos] if frequency is None else int(frequency)
        self.yfrequency[pos] = frequency

    def set_xreverse(self, reverse = None, xside = None):
        pos = self.xside_to_pos(xside)
        direction = self.default.xdirection[pos] if reverse is None else 2 * int(not reverse) - 1
        self.xdirection[pos] = direction

    def set_yreverse(self, reverse = None, yside = None):
        pos = self.yside_to_pos(yside)
        direction = self.default.ydirection[pos] if reverse is None else 2 * int(not reverse) - 1
        self.ydirection[pos] = direction
        
    def set_xaxes(self, lower = None, upper = None):
        self.xaxes[0] = self.default.xaxes[0] if lower is None else bool(lower)
        self.xaxes[1] = self.default.xaxes[1] if upper is None else bool(upper)
        
    def set_yaxes(self, left = None, right = None):
        self.yaxes[0] = self.default.yaxes[0] if left is None else bool(left)
        self.yaxes[1] = self.default.yaxes[1] if right is None else bool(right)

    def set_frame(self, frame = None):
        self.set_xaxes(frame, frame)
        self.set_yaxes(frame, frame)

    def set_grid(self, horizontal = None, vertical = None):
        horizontal = self.default.grid[0] if horizontal is None else bool(horizontal)
        vertical = self.default.grid[1] if vertical is None else bool(vertical)
        self.grid = [horizontal, vertical]

    def set_color(self, color = None):
        color = color if ut.is_color(color) else None
        return self.default.canvas_color if color is None else color

    def set_canvas_color(self, color = None):
        self.canvas_color = self.set_color(color)
        
    def set_axes_color(self, color = None):
        self.axes_color = self.set_color(color)
        
    def set_ticks_color(self, color = None):
        self.ticks_color = self.set_color(color)

    def set_ticks_style(self, style = None):
        style = style if ut.is_style(style) else None
        style = self.default.ticks_style if style is None else ut.clean_styles(style)
        self.ticks_style = style

    def set_theme(self, theme = None):
        theme = 'default' if theme is None or theme not in ut.themes else theme
        self._set_theme(*ut.themes[theme])

    def clear_color(self):
        self.set_theme('clear')

##############################################
#######    Set Functions Utilities    ########
##############################################

    def set_label(self, label = None): 
        label = None if label is None else str(label).strip()
        spaces = ut.only_spaces(label)
        label = None if spaces else label 
        return label

    def correct_xside(self, xside = None): # from axis side to position
        xaxis = self.default.xside
        xside = xaxis[xside - 1] if isinstance(xside, int) and 1 <= xside <= 2 else xaxis[0] if xside is None or xside.strip() not in xaxis else xside.strip()
        return xside

    def correct_yside(self, yside = None):
        yaxis = self.default.yside
        yside = yaxis[yside - 1] if isinstance(yside, int) and 1 <= yside <= 2 else yaxis[0] if yside is None or yside.strip() not in yaxis else yside.strip()
        return yside

    def xside_to_pos(self, xside = None): # from axis side to position
        xside = self.correct_xside(xside)
        pos = self.default.xside.index(xside)
        return pos

    def yside_to_pos(self, yside = None):
        yside = self.correct_yside(yside)
        pos = self.default.yside.index(yside)
        return pos

    def _set_theme(self, canvas_color, axes_color, ticks_color, ticks_style, color_sequence):
        self.canvas_color = canvas_color
        self.axes_color = axes_color
        self.ticks_color = ticks_color
        self.ticks_style = ticks_style
        self.color_sequence = color_sequence
        
##############################################
##########    Draw() Function    #############
##############################################

    def draw(self, *args, **kwargs): # from draw() comes directly the functions scatter() and plot()
        self.add_xside(kwargs.get("xside"))
        self.add_yside(kwargs.get("yside"))
        self.add_data(*args)
        self.add_lines(kwargs.get("lines"))
        self.add_markers(kwargs.get("marker"))
        self.add_colors(kwargs.get("color"))
        self.add_styles(kwargs.get("style"))
        self.add_fillx(kwargs.get("fillx"))
        self.add_filly(kwargs.get("filly"))
        self.add_label(kwargs.get("label"))
        
##############################################
#######    Draw() Called Functions    ########
##############################################

    def add_xside(self, xside = None):
        xside = self.correct_xside(xside)
        self.xside.append(xside)

    def add_yside(self, yside = None):
        yside = self.correct_yside(yside)
        self.yside.append(yside)

    def add_data(self, *args):
        x, y = ut.set_data(*args)
        x, x_date = self.to_time(x)
        y, y_date = self.to_time(y)
        self.x_date[self.xside_to_pos(self.xside[-1])] = x_date
        self.y_date[self.yside_to_pos(self.yside[-1])] = y_date
        self.x.append(x)
        self.y.append(y)
        self.signals += 1

    def add_lines(self, lines):
        lines = self.default.lines if lines is None else bool(lines) 
        self.lines.append(lines)
        
    def add_markers(self, marker = None):
        single_marker = isinstance(marker, str) or marker is None
        marker = self.check_marker(marker) if single_marker else list(map(self.check_marker, marker))
        length = len(self.x[-1])
        marker = ut.to_list(marker, length)
        self.marker.append(marker)

    def add_colors(self, color = None):
        list_color = isinstance(color, list) 
        color = list(map(self.check_color, color)) if list_color else self.check_color(color)
        length = len(self.x[-1])
        self.past_colors = self.past_colors + [color] if color not in self.past_colors else self.past_colors
        color = ut.to_list(color, length)
        self.color.append(color)

    def add_styles(self, style = None):
        single_style = isinstance(style, str) or style is None
        style = self.check_style(style) if single_style else list(map(self.check_style, style))
        length = len(self.x[-1])
        style = ut.to_list(style, length)
        self.style.append(style)

    def add_fillx(self, fillx = None):
        fillx = self.check_fill(fillx)
        self.fillx.append(fillx)

    def add_filly(self, filly = None):
        filly = self.check_fill(filly)
        self.filly.append(filly)

    def add_label(self, label = None):
        spaces = ut.only_spaces(label)
        label = self.default.label if label is None or spaces else str(label).strip() # strip to remove spaces before and after
        self.label.append(label)
        #figure.subplot.label_show.append(default.label_show)
    
##############################################
######    Draw() Functions Utilities   #######
##############################################

    def to_time(self, data):
        dates = any([isinstance(el, str) for el in data])
        data = self.date.strings_to_time(data) if dates else data
        return data, dates
    
    def check_marker(self, marker = None):
        marker = None if marker is None else str(marker)
        marker = self.default.marker if marker is None else marker
        marker = ut.marker_codes[marker] if marker in ut.marker_codes else marker
        marker = marker if marker in ut.hd_symbols else marker[0]
        return marker

    def check_color(self, color = None):
        color = color if ut.is_color(color) else None
        color = self.next_color() if color is None else color
        return color

    def next_color(self):
        color = ut.difference(self.color_sequence, self.past_colors)
        color = color[0] if len(color) > 0 else self.color_sequence[0]
        return color

    def check_style(self, style = None):
        style = None if style is None else str(style)
        style = style if ut.is_style(style) else ut.no_color
        return style

    def check_fill(self, fill = None):
        fill = self.default.fill if fill is None else fill
        fill = False if isinstance(fill, str) and fill != self.default.fill_internal else fill
        fill = 0 if fill is True else fill
        return fill

##############################################
######    Other Plotting Functions    ########
##############################################

    def draw_bar(self, *args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, offset = None, reset_ticks = None, xside = None, yside = None, label = None):
        x, y = ut.set_data(*args)
        marker = self.default.bar_marker if marker is None else marker
        fill = self.default.bar_fill if fill is None else fill
        width = self.default.bar_width if width is None else width
        width = 1 if width > 1 else 0 if width < 0 else width
        orientation = self.check_orientation(orientation, 1)
        minimum = 0 if minimum is None else minimum 
        offset = 0 if offset is None else offset
        reset_ticks = True if reset_ticks is None else reset_ticks

        x_string = any([type(el) == str for el in x]) # if x are strings
        l = len(x)
        xticks = range(1, l + 1) if x_string else x
        xlabels = x if x_string else map(str, x)
        x = xticks if x_string else x
        x = [el + offset for el in x]
        xbar, ybar = ut.bars(x, y, width, minimum)
        xbar, ybar = [xbar, ybar]  if orientation[0] == 'v' else [ybar, xbar]
        (self.set_xticks(xticks, xlabels, xside) if orientation[0] == 'v' else self.set_yticks(xticks, xlabels, yside)) if reset_ticks else None


        firstbar = min([b for b in range(len(x)) if ybar[b][1] != 0], default = 0) # finds the position of the first non zero bar

        for b in range(len(x)):
            xb = xbar[b]; yb = ybar[b]
            plot_label = label if b == firstbar else None
            plot_color = color if b == 0 else self.color[-1]
            nobar = (yb[1] == 0 and orientation[0] == 'v') or (xb[1] == 0 and orientation[0] == 'h')
            plot_marker = " " if nobar else marker
            plot_color = color if b == 0 else self.color[-1][-1]
            self.draw_rectangle(xb, yb,
                                xside = xside, 
                                yside = yside,
                                lines = True,
                                marker = plot_marker,
                                color = plot_color,
                                fill = fill,
                                label = plot_label)

    def draw_multiple_bar(self, *args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, offset = None, reset_ticks = None, xside = None, yside = None, labels = None):
        x, Y = ut.set_multiple_bar_data(*args)
        ly = len(Y)
        width = self.default.bar_width if width is None else width
        marker = [marker] * ly if marker is None or type(marker) != list else marker
        color = [color] * ly if color is None else color
        labels = [labels] * ly if labels is None else labels
        width = width / ly if ly != 0 else 0
        offset = ut.linspace(-1 / 2 + 1 / (2 * ly), 1 / 2 - 1 / (2 * ly), ly) if ly != 0 else []
        
        for i in range(ly):
            self.draw_bar(x, Y[i],
                          marker = marker[i],
                          color = color[i],
                          fill = fill,
                          width = width,
                          orientation = orientation,
                          minimum = minimum,
                          offset = offset[i],
                          xside = xside,
                          yside = yside,
                          label = labels[i],
                          reset_ticks = reset_ticks)

    def draw_stacked_bar(self, *args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, offset = None, reset_ticks = None, xside = None, yside = None, labels = None):
        x, Y = ut.set_multiple_bar_data(*args)
        ly = len(Y)
        marker = [marker] * ly if marker is None or type(marker) != list else marker
        color = [color] * ly if color is None else color
        labels = [labels] * ly if labels is None else labels
        Y = ut.transpose([ut.cumsum(el) for el in ut.transpose(Y)])
        for i in range(ly - 1, -1, -1):
            self.draw_bar(x, Y[i],
                          xside = xside, 
                          yside = yside,
                          marker = marker[i],
                          color = color[i],
                          fill = fill,
                          width = width,
                          orientation = orientation,
                          label = labels[i],
                          minimum = minimum,
                          reset_ticks = reset_ticks)

    def draw_hist(self, data, bins = None, marker = None, color = None, fill = None, norm = None, width = None, orientation = None, minimum = None, xside = None, yside = None, label = None):
        bins = self.default.hist_bins if bins is None else bins
        norm = False if norm is None else norm
        x, y = ut.hist_data(data, bins, norm)
        self.draw_bar(x, y,
                      xside = xside, 
                      yside = yside,
                      marker = marker,
                      color = color,
                      fill = fill,
                      width = width,
                      orientation = orientation,
                      label = label,
                      minimum = None,
                      reset_ticks = False)

    def draw_candlestick(self, dates, data, colors = None, orientation = None, xside = None, yside = None, label = None):
        orientation = self.check_orientation(orientation, 1)
        markers = ['sd', '│', '─'] #if markers is None else markers
        colors = ['green', 'red'] if colors is None else colors
        x = []; y = []; color = []
        ln = len(dates)
        data = {"Open": [], "Close": [], "High": [], "Low": []} if len(data) == 0 else data
        Open = data["Open"]; Close = data["Close"]; High = data["High"]; Low = data["Low"]
        for i in range(ln):
            d = dates[i]
            o, c, h, l = Open[i], Close[i], High[i], Low[i]
            color = colors[0] if c > o else colors[1]
            m, M = min(o, c), max(o, c)
            lab = label if i == 0 else None
            if orientation in ['v', 'vertical']:
                self.draw([d, d], [M, h], xside = xside, yside = yside, color = color, marker = markers[1], lines = True)
                self.draw([d, d], [l, m], xside = xside, yside = yside, color = color, marker = markers[1], lines = True)
                self.draw([d, d], [m, M], xside = xside, yside = yside, color = color, marker = markers[0], lines = True, label = lab)
            elif orientation in ['h', 'horizontal']:
                self.draw([M, h], [d, d], xside = xside, yside = yside, color = color, marker = markers[2], lines = True)
                self.draw([l, m], [d, d], xside = xside, yside = yside, color = color, marker = markers[2], lines = True)
                self.draw([m, M], [d, d], xside = xside, yside = yside, color = color, marker = markers[0], lines = True, label = lab)

    def draw_box(self, *args, xside = None, yside = None, orientation = None, colors = None, label = None, fill = None, width = None, minimum = None, offset = None, reset_ticks = None, quintuples = None):
        x, y = ut.set_data(*args)
        fill = self.default.bar_fill if fill is None else fill
        width = self.default.bar_width if width is None else width
        width = 1 if width > 1 else 0 if width < 0 else width
        orientation = self.check_orientation(orientation, 1)
        minimum = 0 if minimum is None else minimum
        offset = 0 if offset is None else offset
        reset_ticks = True if reset_ticks is None else reset_ticks
        colors = ['green', 'red'] if colors is None else colors
        quintuples = False if quintuples is None else quintuples

        x_string = any([type(el) == str for el in x]) # if x are strings
        l = len(x)
        xticks = range(1, l + 1) if x_string else x
        xlabels = x if x_string else map(str, x)
        x = xticks if x_string else x
        x = [el + offset for el in x]
        (self.set_xticks(xticks, xlabels, xside) if orientation[0] == 'v' else self.set_yticks(xticks, xlabels, yside)) if reset_ticks else None
        if quintuples:
            # todo: check y is aligned.
            _, _, _, _, _, c, xbar = ut.box(x, y, width, minimum)
            q1, q2, q3, max_, min_ = [], [], [], [], []
            for d in y:
                max_.append(d[0])
                q3.append(d[1])
                q2.append(d[2])
                q1.append(d[3])
                min_.append(d[4])
        else:
            q1, q2, q3, max_, min_, c, xbar = ut.box(x, y, width, minimum)
        markers = ['sd', '│', '─'] #if markers is None else markers
        
        for i in range(l):
            lab = label if i == 0 else None
            color = colors[0]
            mcolor = colors[1]
            d, l, h, m, E, M = c[i], min_[i], max_[i], q1[i], q2[i], q3[i]
            Ew = (M - m) / 30
            if orientation in ['v', 'vertical']:
                self.draw([d, d], [M, h], xside = xside, yside = yside, color = color, marker = markers[1], lines = True)
                self.draw([d, d], [l, m], xside = xside, yside = yside, color = color, marker = markers[1], lines = True)
                self.draw_rectangle(xbar[i], [m, M], xside = xside, yside = yside,
                    lines = True, color = color, fill = fill, marker = markers[0], label = lab)
                self.draw_rectangle(xbar[i], [E, E], xside = xside, yside = yside,
                    lines = True, color = mcolor, fill = fill, marker = markers[2])
                #self.draw([d, d], [m, M], xside = xside, yside = yside, color = color, marker = markers[0], lines = True, label = lab)
                #self.draw(xbar[i], [E, E], xside = xside, yside = yside, color = mcolor, marker = markers[0], lines = False)
            elif orientation in ['h', 'horizontal']:
                self.draw([M, h], [d, d], xside = xside, yside = yside, color = color, marker = markers[2], lines = True)
                self.draw([l, m], [d, d], xside = xside, yside = yside, color = color, marker = markers[2], lines = True)
                self.draw_rectangle([m, M], xbar[i], xside = xside, yside = yside,
                    lines = True, color = color, fill = fill, marker = markers[0], label = lab)
                self.draw_rectangle([E, E], xbar[i], xside = xside, yside = yside,
                    lines = True, color = mcolor, fill = fill, marker = markers[1])
                #self.draw([m, M], [d, d], xside = xside, yside = yside, color = color, marker = markers[0], lines = True, label = lab)
                #self.draw([E, E], [d, d], xside = xside, yside = yside, color = 'red', marker = markers[0], lines = True)
        
##############################################
###########    Plotting Tools    #############
##############################################
        
    def draw_error(self, *args, xerr = None, yerr = None, color = None, xside = None, yside = None, label = None):
        x, y = ut.set_data(*args)
        l = len(x)
        xerr = [0] * l if xerr is None else xerr
        yerr = [0] * l if yerr is None else yerr
        for i in range(l):
            col = self.color[-1][-1] if i > 0 else color
            self.draw([x[i], x[i]], [y[i] - yerr[i] / 2, y[i] + yerr[i] / 2], xside = xside, yside = yside, marker = "│", color = col, lines = True)
            col = self.color[-1][-1] if i == 0 else col
            self.draw([x[i] - xerr[i] / 2, x[i] + xerr[i] / 2], [y[i], y[i]], xside = xside, yside = yside, marker = "─", color = col, lines = True)
            self.draw([x[i]], [y[i]], xside = xside, yside = yside, marker = "┼", color = col, lines = True)

    def draw_event_plot(self, data, marker = None, color = None, orientation = None, side = None):
        x, y = data, [1.1] * len(data)
        orientation = self.check_orientation(orientation, 1)
        if orientation in ['v', 'vertical']:
            self.draw(x, y, xside = side, marker = marker, color = color, fillx = True)
            self.set_ylim(0, 1)
            self.set_yfrequency(0)
        else:
            self.draw(y, x, yside = side, marker = marker, color = color, filly = True)
            self.set_xlim(0, 1)
            self.set_xfrequency(0)

    def draw_vertical_line(self, coordinate, color = None, xside = None):
        coordinate = self.date.string_to_time(coordinate) if isinstance(coordinate, str) else coordinate
        pos = self.xside_to_pos(xside)
        self.vcoord[pos].append(coordinate)
        color = self.ticks_color if color is None else color
        self.vcolors[pos].append(self.check_color(color))

    def draw_horizontal_line(self, coordinate, color = None, yside = None):
        coordinate = self.date.string_to_time(coordinate) if isinstance(coordinate, str) else coordinate
        pos = self.xside_to_pos(yside)
        self.hcoord[pos].append(coordinate)
        color = self.ticks_color if color is None else color
        self.hcolors[pos].append(self.check_color(color))

    def draw_text(self, text, x, y, xside = None, yside = None, color = None, background = None, style = None, orientation = None, alignment = None):
        orientation = self.check_orientation(orientation)
        text = text if orientation is self.default.orientation[0] else text[::-1]
        self.text.append(str(text))
        x = self.date.string_to_time(x) if isinstance(x, str) else x
        y = self.date.string_to_time(y) if isinstance(y, str) else y
        self.tx.append(x)
        self.ty.append(y) 
        self.txside.append(self.correct_xside(xside))
        self.tyside.append(self.correct_yside(yside))
        color = self.next_color() if color is None or not ut.is_color(color) else color
        background = self.canvas_color if background is None or not ut.is_color(background) else background
        self.tfull.append(color)
        self.tback.append(background)
        self.tstyle.append(self.check_style(style))
        alignment = self.check_alignment(alignment)
        self.torien.append(orientation)
        self.talign.append(alignment)

    def draw_rectangle(self, x = None, y = None, marker = None, color = None, lines = None, fill = None, reset_lim = False, xside = None, yside = None, label = None):
        x = [0, 1] if x is None or len(x) < 2 else x  
        y = [0, 1] if y is None or len(y) < 2 else y  
        xpos = self.xside_to_pos(xside)
        ypos = self.yside_to_pos(yside)
        lines = True if lines is None else lines
        fill = False if fill is None else fill
        xm = min(x); xM = max(x);
        ym = min(y); yM = max(y);
        dx = abs(xM - xm); dy = abs(yM - ym);
        if reset_lim:
            self.xlim[xpos] = [xm - 0.5 * dx, xM + 0.5 * dx]
            self.ylim[xpos] = [ym - 0.5 * dy, yM + 0.5 * dy]
        x, y = [xm, xm, xM, xM, xm], [ym, yM, yM, ym, ym]
        self.draw(x, y,
                  xside = xside, 
                  yside = yside,
                  lines = True if fill else lines,
                  marker = marker,
                  color = color,
                  fillx = "internal" if fill else False,
                  filly = False,
                  label = label)

    def draw_polygon(self, x = None, y = None,  radius = None, sides = None, marker = None, color = None, lines = None, fill = None, reset_lim = False, xside = None, yside = None, label = None):
        x = 0 if x is None else x
        y = 0 if y is None else y
        radius = 1 if radius is None else abs(int(radius))
        sides = 3 if sides is None else max(3, int(abs(sides)))
        xpos = self.xside_to_pos(xside)
        ypos = self.yside_to_pos(yside)
        lines = True if lines is None else lines
        fill = False if fill is None else fill
        
        alpha = 2 * math.pi / sides
        init = alpha / 2 + math.pi / 2 if sides % 2 == 0 else alpha / 4 * ((-1) ** (sides // 2))# * math.pi #- ((-1) ** (sides)) * alpha / 4
        #init = 0 * init
        get_point = lambda i: [x + math.cos(alpha * i + init) * radius, y + math.sin(alpha * i + init) * radius]
        # the rounding is needed so that results like 9.9999 are rounded to 10 and display as same coordinate in the plot, otherwise the floor function will turn 9.999 into 9
        points = [get_point(i) for i in range(sides + 1)]
        if reset_lim:
            self.xlim[xpos] = [x - 1.5 * radius, x + 1.5 * radius]
            self.ylim[xpos] = [y - 1.5 * radius, y + 1.5 * radius]
        self.draw(*ut.transpose(points),
                  xside = xside, 
                  yside = yside,
                  lines = True if fill else lines,
                  marker = marker,
                  color = color,
                  fillx = "internal" if fill else False,
                  filly = False,
                  label = label)
        
    def draw_confusion_matrix(self, actual, predicted, color = None, style = None, labels = None):
        color = self.default.cmatrix_color if color is None else self.check_color(color)
        style = self.default.cmatrix_style if style is None else self.check_style(style)
        
        L = len(actual)
        n_labels = sorted(ut.no_duplicates(actual))
        labels = n_labels if labels is None else list(labels)
        l = len(n_labels)

        get_sum = lambda a, p: sum([actual[i] == a and predicted[i] == p for i in range(L)])
        cmatrix = [[get_sum(n_labels[r], n_labels[c]) for c in range(l)] for r in range(l)]
        cm = ut.join(cmatrix); m, M, t = min(cm), max(cm), sum(cm)

        lm = 253; lM = 80
        to_255 = lambda l: round(lm + (lM - lm) * (l - m) / (M - m)) # l=m -> lm; l=M->lM
        to_color = lambda l: tuple([to_255(l)] * 3)
        to_text = lambda n: str(round(n, 2)) + ' - ' + str(round(100 * n / t, 2)) + '%'
        for r in range(l):
            for c in range(l):
                count = cmatrix[r][c]
                col = to_color(count)
                self.draw_rectangle([c - 0.5, c + 0.5], [r - 0.5, r + 0.5], color = col, fill = True)
                self.draw_text(to_text(count), c, r, color = color, background = col, style = style)

        self.set_yreverse(True)
        self.set_xticks(n_labels, labels)
        self.set_yticks(n_labels, labels)
        self.set_ticks_color(color); self.set_ticks_style(style);
        self.set_axes_color('default'); self.set_canvas_color('default');
        self.set_title('Confusion Matrix')
        self.set_xlabel('Predicted')
        self.set_ylabel('Actual')

    def draw_indicator(self, value, label = None, color = None, style = None):
        color = self.default.cmatrix_color if color is None else self.check_color(color)
        style = self.default.cmatrix_style if style is None else self.check_style(style)

        self.set_title(label)
        self.set_ticks_color(color);
        self.set_ticks_style(style);
        self.set_axes_color('default')
        self.set_canvas_color('default')
        self.set_xfrequency(0)
        self.set_yfrequency(0)

        self.draw_text(str(value), 0, 0, color = color, style = style, alignment = 'center')

##############################################
##############    2D Plots    ################
############################################## 
    
    def draw_matrix(self, matrix, marker = None, style = None, fast = False):
        matrix = [l.copy() for l in matrix]
        marker = [marker] if type(marker) != list else marker
        marker = [self.check_marker("sd") if el in ut.join([None, ut.hd_symbols]) else self.check_marker(el) for el in marker]
        style = ut.no_color if style is None else self.check_style(style)
        cols, rows = ut.matrix_size(matrix)
        rows = 0 if cols == 0 else rows
        matrix = matrix if rows * cols != 0 and ut.is_rgb_color(matrix[0][0]) else ut.turn_gray(matrix)
        marker = ut.repeat(marker, cols)
        if not fast:
            for r in range(rows):
                xyc = [(c, r, matrix[rows - 1 - r][c]) for c in range(cols)]
                x, y, color = ut.transpose(xyc, 3)
                self.draw(x, y, marker = marker, color = color, style = style)
            self.set_canvas_color("black")
            self.set_xlabel('column') 
            self.set_ylabel('row')
            xf, yf = min(self.xfrequency[0], cols), min(self.yfrequency[0], rows)
            xt = ut.linspace(0, cols - 1, xf)
            xl = ut.get_labels([el + 1 for el in xt])
            yt = ut.linspace(0, rows - 1, yf)
            yl = ut.get_labels([rows - el for el in yt])
            self.set_xticks(xt, xl)
            self.set_yticks(yt, yl)
        else: # if fast
            for r in range(rows):
                for c in range(cols):
                    ansi = ut.colors_to_ansi(matrix[r][c], style, "black")
                    matrix[r][c] = ansi + marker[c] + ut.ansi_end
            self.matrix.canvas = '\n'.join([''.join(row) for row in matrix])
            self.fast_plot = True

    def draw_heatmap(self, dataframe, color = None, style=None):
        color = self.default.cmatrix_color if color is None else self.check_color(color)
        style = self.default.cmatrix_style if style is None else self.check_style(style)

        xlabels = dataframe.columns.tolist()
        ylabels = dataframe.index.tolist()

        cmatrix = dataframe.values.tolist()
        cm = ut.join(cmatrix)
        m, M, t = min(cm), max(cm), sum(cm)

        lm = 253
        lM = 80
        to_255 = lambda l: round(lm + (lM - lm) * (l - m) / (M - m))  # l=m -> lm; l=M->lM
        to_color = lambda l: tuple([to_255(l)] * 3)

        for r in range(len(dataframe.index.tolist())):
            for c in range(len(dataframe.columns.tolist())):
                count = cmatrix[r][c]
                col = to_color(count)
                self.draw_rectangle([c - 0.5, c + 0.5], [r - 0.5, r + 0.5], marker= 'sd', color=col, fill=True)

        y_labels = list(set(range(len(dataframe.columns))))
        x_labels = list(set(range(len(dataframe.columns))))

        self.set_yreverse(True)
        self.set_xticks(x_labels, xlabels)
        self.set_yticks(y_labels, ylabels)
        self.set_ticks_color(color);
        self.set_ticks_style(style);
        self.set_axes_color('default');
        self.set_canvas_color('default');
        self.set_title('Heatmap')
        print(dataframe)

    def draw_image(self, path, marker = None, style = None, fast = False, grayscale = False):
        from PIL import Image        
        path = ut.correct_path(path)
        if not ut.is_file(path):
            return
        image = Image.open(path)
        self._draw_image(image, marker = marker, style = style, grayscale = grayscale, fast = fast)
        
##############################################
#######    Plotting Tools Utilities    #######
##############################################

    def check_orientation(self, orientation = None, default_index = 0):
        default = self.default.orientation
        default_first_letter = [el[0] for el in default]
        orientation = default[default_first_letter.index(orientation)] if orientation in default_first_letter else orientation
        orientation = default[default_index] if orientation not in default else orientation
        return orientation

    def check_alignment(self, alignment = None):
        default = self.default.alignment[0:-1]
        default_first_letter = [el[0] for el in default]
        alignment = default[default_first_letter.index(alignment)] if alignment in default_first_letter else alignment
        alignment = default[1] if alignment not in default else alignment
        return alignment

    def _draw_image(self, image, marker = None, style = None, fast = False, grayscale = False):
        from PIL import ImageOps
        image = ImageOps.grayscale(image) if grayscale else image
        image = image.convert('RGB')
        size = ut.update_size(image.size, self.size)
        image = image.resize(size, resample = True)
        matrix = ut.image_to_matrix(image)
        self.set_xfrequency(0); self.set_yfrequency(0);
        self.draw_matrix(matrix, marker = marker, style = style, fast = fast)
        self.set_xlabel(); self.set_ylabel()


import plotext._utility as ut
import math

# this file builds a class inherited by the monitor_class() in _monitor.py just because its only method - build_plot() - is very long and it is the core of plot building and it is written separately for clarity

class build_class():
    
    def build_plot(self): # it builds the plot given the external settings and internal settings collected in draw()

        # Initial Tools
        r2 = [0, 1]
        signals = len(self.x); Signals = list(range(signals))
        texts = len(self.text); Texts = list(range(texts))
        width, height = self.size
        ticks_colors = self.ticks_color, self.ticks_style

        # Find if Axes are used
        xside = [(self.default.xside[i] in self.xside + self.txside) or self.vcoord[i] != [] for i in r2]
        yside = [(self.default.yside[i] in self.yside + self.tyside) or self.hcoord[i] != [] for i in r2]

        # Remove Useless X and Y Ticks and Labels if axes are not used 
        self.xticks = [self.xticks[i] if xside[i] else None for i in r2]
        self.xlabels = [self.xlabels[i] if xside[i] else None for i in r2]

        self.yticks = [self.yticks[i] if yside[i] else None for i in r2]
        self.ylabels = [self.ylabels[i] if yside[i] else None for i in r2]

        # Remove Useless h and v user defined Lines if axes are not used
        self.hcoord = [self.hcoord[i] if yside[i] else [] for i in r2]
        self.vcoord = [self.vcoord[i] if xside[i] else [] for i in r2]
        self.hcolors = [self.hcolors[i] if yside[i] else [] for i in r2]
        self.vcolors = [self.vcolors[i] if xside[i] else [] for i in r2]

        # Apply Scale (linear or log) to the data
        xscale = [ut.get_first(self.xscale, self.xside[s] is self.default.xside[0]) for s in Signals] # the x scale for each signal 
        yscale = [ut.get_first(self.yscale, self.yside[s] is self.default.yside[0]) for s in Signals] # the y scale for each signal 
        self.x = [ut.apply_scale(self.x[s], xscale[s] is self.default.xscale[1]) for s in Signals] # apply optional log scale
        self.y = [ut.apply_scale(self.y[s], yscale[s] is self.default.yscale[1]) for s in Signals]

        # Apply Scale (linear or log) to the Axes Ticks
        self.xticks = [ut.apply_scale(self.xticks[i], self.xscale[i] is self.default.xscale[1]) if self.xticks[i] is not None else None for i in r2] # apply optional log scale
        self.yticks = [ut.apply_scale(self.yticks[i], self.yscale[i] is self.default.yscale[1]) if self.yticks[i] is not None else None  for i in r2] # apply optional log scale
        
        # Apply Scale (linear or log) to the user defined Lines
        self.hcoord = [ut.apply_scale(self.hcoord[i], self.yscale[i] is self.default.yscale[1]) for i in r2] # apply optional log scale
        self.vcoord = [ut.apply_scale(self.vcoord[i], self.xscale[i] is self.default.xscale[1]) for i in r2] # apply optional log scale

        # Apply Scale (linear or log) to the user defined Text
        txscale = [ut.get_first(self.xscale, self.txside[s] is self.default.xside[0]) for s in Texts] # the x scale for each text
        tyscale = [ut.get_first(self.yscale, self.tyside[s] is self.default.yside[0]) for s in Texts] # the y scale for each text 
        self.tx = [ut.apply_scale([self.tx[s]], txscale[s] is self.default.xscale[1])[0] for s in Texts] #if width_canvas * height_canvas > 0 else [] # apply optional log scale
        self.ty = [ut.apply_scale([self.ty[s]], tyscale[s] is self.default.yscale[1])[0] for s in Texts] #if width_canvas * height_canvas > 0 else [] # apply optional log scale
        tx = [[self.tx[s] for s in Texts if self.txside[s] is self.default.xside[i]] for i in r2] # text x coord for each axis
        ty = [[self.ty[s] for s in Texts if self.tyside[s] is self.default.yside[i]] for i in r2] # text x coofor each axis

        # Get X Axes Limits
        x = [ut.join([self.x[s] for s in Signals if self.xside[s] is side]) for side in self.default.xside] # total x data for each axis
        x = [x[i] + self.vcoord[i] + tx[i] for i in r2] # add v lines and text coords to calculate xlim
        xlim = [ut.get_lim(el) if len(el) > 0 else [None, None] for el in x]
        self.xlim = [ut.replace_none(self.xlim[i], xlim[i]) for i in r2]
        self.xlim = [self.xlim[i][:: self.xdirection[i]] for i in r2] # optionally reverse axes
        xlim = [self.xlim[0] if self.xside[s] == self.default.xside[0] else self.xlim[1] for s in Signals] # xlim for each signal

        # Get Y Axes Limits
        y = [ut.join([self.y[s] for s in Signals if self.yside[s] is side]) for side in self.default.yside]
        y = [y[i] + self.hcoord[i] + ty[i] for i in r2] # add h lines and text coords to calculate ylim
        ylim = list(map(ut.get_lim, y))
        self.ylim = [ut.replace_none(self.ylim[i], ylim[i]) for i in r2]
        self.ylim = [self.ylim[i][:: self.ydirection[i]] for i in r2] # optionally reverse axes
        ylim = [self.ylim[0] if self.yside[s] == self.default.yside[0] else self.ylim[1] for s in Signals] # ylim for each signal

        # Get Y Ticks and Labels
        yticks_to_set = [self.yticks[i] is None and yside[i] and len(y[i]) > 0 for i in r2]
        yticks = [ut.linspace(*self.ylim[i], self.yfrequency[i]) if yticks_to_set[i] else self.yticks[i] for i in r2] # the actual Y ticks
        yticks_rescaled = [ut.reverse_scale(yticks[i], self.yscale[i] is self.default.yscale[1]) for i in r2]
        
        ylabels = [self.date.times_to_string(yticks_rescaled[i]) if self.y_date[i] else ut.get_labels(yticks_rescaled[i]) if yticks_to_set[i] else self.ylabels[i] for i in r2]
        ylabels = [ut.add_extra_spaces(ylabels[i], self.default.yside[i]) if ylabels[i] is not None else None for i in r2]
        width_ylabels = [ut.max_length(el) if el is not None else 0 for el in ylabels]

        # Get X Ticks and Labels
        xticks_to_set = [self.xticks[i] is None and xside[i] and len(x[i]) > 0 for i in r2]
        xticks = [ut.linspace(*self.xlim[i], self.xfrequency[i]) if xticks_to_set[i] else self.xticks[i] for i in r2] # the actual X ticks
        xticks_rescaled = [ut.reverse_scale(xticks[i], self.xscale[i] is self.default.xscale[1]) for i in r2]
        xlabels = [self.date.times_to_string(xticks_rescaled[i]) if self.x_date[i] else ut.get_labels(xticks_rescaled[i]) if xticks_to_set[i] else self.xlabels[i] for i in r2]
        xlabels = [ut.add_extra_spaces(xlabels[i], self.default.xside[i]) if xticks_to_set[i] else self.xlabels[i] for i in r2]
        height_xlabels = [len(el) > 0 if el is not None else 0 for el in xlabels]

        # Canvas Dimensions (the area of data points)
        width_canvas = width - sum(self.yaxes) - sum(width_ylabels)
        height_highbar = any([el is not None for el in [self.title, self.xlabel[1]]])
        height_lowbar = any([el is not None for el in self.ylabel + [self.xlabel[0]]])
        height_canvas = height - sum(self.xaxes) - sum(height_xlabels) - height_highbar - height_lowbar
        
        # Canvas Offset
        col_start = width_ylabels[0] + self.yaxes[0]
        col_end = col_start + width_canvas
        row_start = height_lowbar + height_xlabels[0] + self.xaxes[0]
        row_end = row_start + height_canvas

        # Get Absolute X and Y Ticks
        cticks = [ut.get_matrix_data(xticks[i], self.xlim[i], width_canvas) if xticks[i] != None else [] for i in r2]
        rticks = [ut.get_matrix_data(yticks[i], self.ylim[i], height_canvas) if yticks[i] != None else [] for i in r2]

        # Get Absolute Coordinates for user defined Lines 
        hticks = [ut.get_matrix_data(self.hcoord[i], self.ylim[i], height_canvas) if None not in self.ylim[i] else [] for i in r2]
        vticks = [ut.get_matrix_data(self.vcoord[i], self.xlim[i], width_canvas) if None not in self.xlim[i] else [] for i in r2]

        # Get Absolute Coordinates for user defined Text
        txlim = [ut.get_first(self.xlim, self.txside[s] == self.default.xside[0]) for s in Texts] # xlim for each text
        tylim = [ut.get_first(self.ylim, self.tyside[s] == self.default.yside[0]) for s in Texts] # xlim for each text
        tcticks = [ut.get_matrix_data([self.tx[s]], txlim[s], width_canvas)[0] if width_canvas > 0 else 0 for s in Texts] #if width_canvas > 0 else [] # 
        trticks = [ut.get_matrix_data([self.ty[s]], tylim[s], height_canvas)[0] if height_canvas > 0 else 0  for s in Texts] #if height_canvas > 0 else [] #

        # Get effective number of User Lines
        hlines = [len(el) for el in hticks]; Hlines = [list(range(el)) for el in hlines] # number of user defined horizontal lines for each y axis (as list)
        vlines = [len(el) for el in vticks]; Vlines = [list(range(el)) for el in vlines] # number of user defined vertical lines for each x axis (as list)

        # Create Matrix of Markers
        self.matrix.set_size(width, height)
        self.matrix.set_axes_color(self.axes_color)
        self.matrix.set_canvas_area(col_start, col_end, row_start, row_end)
        self.matrix.set_canvas_color(self.canvas_color)
        self.matrix.set_matrices()

        # Add Title
        col_center = col_start + width_canvas // 2 # centered on the canvas not the entire plot
        col_title = col_center if self.xlabel[1] is None else 0
        row_title = row_end + self.xaxes[1] + height_xlabels[1]
        alignment_title = "center" if self.xlabel[1] is None else "left"
        self.matrix.add_horizontal_string(col_title, row_title, self.title, *ticks_colors, alignment = alignment_title, check_space = True) if self.title and height > 0 else None

        # Add Upper X Label
        self.matrix.add_horizontal_string(col_center, row_title, self.xlabel[1], *ticks_colors, alignment = "center", check_space = True) if self.xlabel[1] and height > 0 else None

        # Add Lower X Ticks
        row_xticks = min(row_start - self.xaxes[0] - 1, height - 1)
        cticks_inserted = [height > 0 and self.matrix.add_horizontal_string(col_start + cticks[0][i], row_xticks, xlabels[0][i], *ticks_colors, alignment = "dynamic", check_space = True) for i in range(len(cticks[0]))]
        cticks[0] = [cticks[0][i] for i in range(len(cticks[0])) if cticks_inserted[i]] # it updates the x ticks coordinates whatever x labels were inserted 

        # Add Upper X Ticks: the reason to do it here prematurely, is that the cticks[1] need to be re-evaluated for next step 
        row_xticks = row_end + self.xaxes[1]
        cticks_inserted = [height > 0 and self.matrix.add_horizontal_string(col_start + cticks[1][i], row_xticks, xlabels[1][i], *ticks_colors, alignment = "dynamic", check_space = True) for i in range(len(cticks[1]))]
        cticks[1] = [cticks[1][i] for i in range(len(cticks[1])) if cticks_inserted[i]] # it updates the x ticks coordinates whatever x labels were inserted 

        # Add Upper X Axes (from previous step)
        tick = lambda i: '┼' if (self.grid[1] and i in cticks[1]) else '┬' if (self.grid[1] and i in cticks[0] or i in ut.join(vticks)) else '┴' if i in cticks[1] else '─'
        xaxis = [tick(i) for i in range(width_canvas)]
        self.matrix.add_horizontal_string(col_start, row_end, xaxis, self.ticks_color) if self.xaxes[0] and height_canvas >= -1 else None   

        # Add Left Y Ticks
        [self.matrix.add_horizontal_string(0, rticks[0][i] + row_start, ylabels[0][i], *ticks_colors) for i in range(len(rticks[0]))] if width >= width_ylabels[0] else None
        
        # Add Left Y Axis
        tick = lambda i: '┼' if (self.grid[0] and i in rticks[0]) else '├' if (self.grid[0] and i in rticks[1] or i in ut.join(hticks)) else '┤' if i in rticks[0] else '│'
        yaxis = [tick(i) for i in range(height_canvas)]
        col_yaxis = width_ylabels[0]
        self.matrix.add_vertical_string(col_yaxis, row_start, yaxis, self.ticks_color) if self.yaxes[0] and width >= sum(width_ylabels) + 1 else None

        # Add Right Y Axis
        tick = lambda i: '┼' if (self.grid[0] and i in rticks[1]) else '┤' if (self.grid[0] and i in rticks[0] or i in ut.join(hticks)) else '├' if i in rticks[1] else '│'
        yaxis = [tick(i) for i in range(height_canvas)]
        self.matrix.add_vertical_string(col_end, row_start, yaxis, self.ticks_color) if self.yaxes[1] and width >= sum(width_ylabels) + self.yaxes[0] + 1 else None

        # Add Right Y Ticks
        [self.matrix.add_horizontal_string(col_end + 1, rticks[1][i] + row_start, ylabels[1][i], *ticks_colors) for i in range(len(rticks[1]))] if width >= sum(width_ylabels) + 1 else None

        # Add Frame 4 Corners if necessary
        canvas_test = height_canvas >= 0 and width_canvas >= 0
        self.matrix.insert_element(col_start - 1, row_start - 1, '└', self.ticks_color) if self.xaxes[0] and self.yaxes[0] and canvas_test else None
        self.matrix.insert_element(col_end, row_start - 1, '┘', self.ticks_color) if self.xaxes[0] and self.yaxes[1] and canvas_test else None
        self.matrix.insert_element(col_start - 1, row_end, '┌', self.ticks_color) if self.xaxes[1] and self.yaxes[0] and canvas_test else None
        self.matrix.insert_element(col_end, row_end, '┐', self.ticks_color) if self.xaxes[1] and self.yaxes[1] and canvas_test else None

        # Add Lower X Axes (from previous step)
        tick = lambda i: '┼' if (self.grid[1] and i in cticks[0]) else '┴' if (self.grid[1] and i in cticks[1] or i in ut.join(vticks)) else '┬' if i in cticks[0] else '─'
        xaxis = [tick(i) for i in range(width_canvas)]
        self.matrix.add_horizontal_string(col_start, row_start - 1, xaxis, self.ticks_color) if self.xaxes[0] and height_canvas >= -1 else None

        # Add Left Y Label
        self.matrix.add_horizontal_string(0, 0, self.ylabel[0], *ticks_colors, check_space = True) if self.ylabel[0] and height > 0 else None
        
        # Add Right Y Label 
        self.matrix.add_horizontal_string(width - 1, 0, self.ylabel[1], *ticks_colors, check_space = True, alignment = "right") if self.ylabel[1] and height > 0 else None

        # Add Lower X Label 
        self.matrix.add_horizontal_string(col_center, 0, self.xlabel[0], *ticks_colors, alignment = "center", check_space = True) if self.xlabel[0] and height > 0 else None

        # Add Grid Lines
        hline = '─' * width_canvas
        [self.matrix.add_horizontal_string(0 + col_start, row + row_start, hline, self.ticks_color) for row in ut.join(rticks) if self.grid[0]]
        vline = '│' * height_canvas
        [self.matrix.add_vertical_string(col + col_start, 0 + row_start, vline, self.ticks_color)  for col in ut.join(cticks) if self.grid[1]]
        [self.matrix.insert_element(col + col_start, row + row_start, '┼') for row in ut.join(rticks) for col in ut.join(cticks) if all(self.grid)] # deals with the crossing between grids
        
        # Add user defined Lines
        [[self.matrix.add_horizontal_string(col_start, hticks[i][l] + row_start, hline, self.hcolors[i][l]) for l in Hlines[i]] for i in r2]
        [[self.matrix.add_vertical_string(vticks[i][l] + col_start, 0 + row_start, vline, self.vcolors[i][l]) for l in Vlines[i]] for i in r2]
        [[[self.matrix.insert_element(col + col_start, hticks[i][l] + row_start, '┼', self.hcolors[i][l]) for l in Hlines[i]] for i in r2] for col in ut.join(cticks) if self.grid[1]] # deals with the crossing between h lines and v grids
        [[[self.matrix.insert_element(vticks[i][l] + col_start, row + row_start, '┼', self.vcolors[i][l]) for l in Vlines[i]] for i in r2] for row in ut.join(rticks) if self.grid[0]] # deals with the crossing between v lines and h grids
        [[[[self.matrix.insert_element(col_start + vticks[iv][lv], hticks[i][l] + row_start, '┼', self.hcolors[i][l]) for l in Hlines[i]] for i in r2] for lv in Vlines[iv]] for iv in r2] # deals with the crossing between h and v lines

        # Expand Canvas to accommodate HD markers
        xf = [max([ut.marker_factor(el, 2, 2, 2) for el in self.marker[s]], default = 1) for s in Signals]
        yf = [max([ut.marker_factor(el, 2, 3, 4) for el in self.marker[s]], default = 1) for s in Signals]
        width_expanded = [width_canvas * el for el in xf]
        height_expanded = [height_canvas * el for el in yf]

        # Get Relative Data to Be Plotted on Matrix
        test_canvas = [width_expanded[s] * width_expanded[s] for s in Signals]
        x = [ut.get_matrix_data(self.x[s], xlim[s], width_expanded[s]) if test_canvas[s] else [] for s in Signals]
        y = [ut.get_matrix_data(self.y[s], ylim[s], height_expanded[s]) if test_canvas[s] else []  for s in Signals]
        m, c, st = self.marker, self.color, self.style


        # Add Lines between Data Points
        x, y, m, c, st = ut.transpose([ut.get_lines(x[s], y[s], m[s], c[s], st[s]) if self.lines[s] else (x[s], y[s], m[s], c[s], st[s]) for s in Signals], 5)

        # Fillx
        #x, y, m, c, st = ut.transpose([ut.brush(x[s], y[s], m[s], c[s], st[s]) for s in Signals], 5)
        level = [ut.get_fill_level(self.fillx[s], ylim[s], height_expanded[s]) for s in Signals]
        x, y, m, c, st = ut.transpose([ut.fill_data(x[s], y[s], level[s], m[s], c[s], st[s]) if self.fillx[s] is not False else (x[s], y[s], m[s], c[s], st[s]) for s in Signals], 5)
        
        # Filly
        #x, y, m, c, st = ut.transpose([ut.brush(x[s], y[s], m[s], c[s], st[s]) for s in Signals], 5)
        level = [ut.get_fill_level(self.filly[s], xlim[s], width_expanded[s]) for s in Signals]
        y, x, m, c, st = ut.transpose([ut.fill_data(y[s], x[s], level[s], m[s], c[s], st[s]) if self.filly[s] is not False else (y[s], x[s], m[s], c[s], st[s]) for s in Signals], 5)

        # Get Actual HD Markers
        x, y, m, c, st = ut.transpose([ut.brush(x[s], y[s], m[s], c[s], st[s]) for s in Signals], 5)
        xf = [[ut.marker_factor(el, 2, 2, 2) for el in m[s]] for s in Signals]
        yf = [[ut.marker_factor(el, 2, 3, 4) for el in m[s]] for s in Signals]
        test = [max(xf[s], default = 1) * max(yf[s], default = 1) != 1 for s in Signals]
        x, y, mxy = ut.transpose([ut.hd_group(x[s], y[s], xf[s], yf[s]) if test[s] else (x[s], y[s], []) for s in Signals], 3)
        m = [[ut.get_hd_marker(mxy[s][i]) if m[s][i] in ut.hd_symbols else m[s][i] for i in range(len(x[s]))] for s in Signals]

        # Add Data to Canvas
        x, y, m, c, st = ut.transpose([ut.remove_outsiders(x[s], y[s], width_canvas, height_canvas, m[s], c[s], st[s]) for s in Signals], 5)

        x, y, m, c, st = ut.transpose([ut.brush(x[s], y[s], m[s], c[s], st[s]) for s in Signals], 5)
        [[self.matrix.insert_element(x[s][i] + col_start, y[s][i] + row_start, m[s][i], c[s][i], st[s][i]) for i in range(len(x[s]))] for s in Signals]

        # Legend Utilities
        labelled = lambda s: self.label[s] is not None
        labels = [ut.space + self.label[s] + ut.space for s in Signals if labelled(s)]; l = len(labels); L = ut.max_length(labels)
        labels = [el + ut.space * (L - len(el)) for el in labels]

        # Add Legend Side Symbols
        side = [ut.space + ut.side_symbols[(self.xside[s], self.yside[s])] for s in Signals if labelled(s)]
        side_test = not (ut.no_duplicates(side) == [' L']) and not l == 1; S = 2 if side_test else 0;
        legend_test = width_canvas >= S + 3 + L and height_canvas >= len(labels)
        [self.matrix.add_horizontal_string(col_start, row_end - 1 - s, side[s], self.ticks_color, self.ticks_style) for s in range(l)] if legend_test and side_test else None

        # Add Legend Markers
        take_3 = lambda data: (data[ : 3] * 3)[ : 3]
        marker = [take_3(self.marker[s]) for s in Signals if labelled(s)]
        replace_hd_marker = lambda marker: ut.hd_symbols[marker] if marker in ut.hd_symbols else marker
        marker = [[ut.space] + list(map(replace_hd_marker, el)) for el in marker]
        color =  [[ut.no_color] + take_3(c[s]) for s in Signals if labelled(s)]
        style =  [[ut.no_color] + take_3(st[s]) for s in Signals if labelled(s)]
        [[self.matrix.insert_element(col_start + S + i, row_end - 1 - s, marker[s][i], color[s][i], style[s][i]) for i in range(3)] for s in range(l)] if legend_test else None
        [self.matrix.add_horizontal_string(col_start + S + 3, row_end - 1 - s, labels[s], self.ticks_color, self.ticks_style) for s in range(l)] if legend_test else None

        # Add Text to Canvas
        [self.matrix.add_multiple_horizontal_strings(col_start + tcticks[s], row_start + trticks[s], self.text[s], self.tfull[s], self.tstyle[s], self.tback[s], self.talign[s], False, True) for s in Texts if self.torien[s] is self.default.orientation[0]] 
        [self.matrix.add_multiple_vertical_strings(col_start + tcticks[s], row_start + trticks[s], self.text[s], self.tfull[s], self.tstyle[s], self.tback[s], self.talign[s], True) for s in Texts if self.torien[s] is self.default.orientation[1]]







#from plotext._utility import bar_marker
from plotext._utility import no_color, plot_marker

class default_figure_class():
    
    def __init__(self):
        self.set_limitsize()
        self.set_size_term()
        self.set_size_term_inf()
        self.interactive = False

    def set_limitsize(self, limit_width = None, limit_height = None):
        self.limit_width = True if limit_width is None else bool(limit_width)
        self.limit_height = True if limit_height is None else bool(limit_height)
        self.limit_size = [self.limit_width, self.limit_height]

    def set_size_term(self, width = None, height = None):
        self.width_term = 211 * 2 // 3 if width is None else int(width)
        self.height_term = 53 * 2 // 3 if height is None else int(height)
        self.size_term = [self.width_term, self.height_term]

    def set_size_term_inf(self, width = None, height = None):
        m = 5
        self.width_term_inf = m * self.width_term if width is None else int(width)
        self.height_term_inf = m * self.height_term if height is None else int(height)
        self.size_term_inf = [self.width_term, self.height_term]

        
class default_monitor_class():
    
    def __init__(self):
        self.marker = plot_marker
        self.color_init()
        self.axes_init()
        self.canvas_init()
        self.text_init()
        self.draw_init()
        self.bar_init()
        self.confusion_matrix_init()

    def color_init(self): # Default Values for Color Set with Outside Functions
        self.canvas_color = "white"
        self.axes_color = "white"
        self.ticks_color = "black"
        self.ticks_style = no_color

    def axes_init(self):  # Default Values for Variables Set with Outside Functions
        self.xaxes = [True, True]
        self.yaxes = [True, True]
        self.xfrequency = [5, 5] # lower and upper xaxes ticks frequency
        self.yfrequency = [7, 7] # left and right yaxes ticks frequency
        self.xdirection = [1, 1] # direction of x axes
        self.ydirection = [1, 1]
        self.xticks = [None, None] # xticks coordinates for both axes
        self.yticks = [None, None]

    def canvas_init(self):
        self.xscale = ["linear", "log"] # the two possibilities, the first is default
        self.yscale = ["linear", "log"]
        self.grid = [False, False]
    
    def text_init(self):
        self.alignment = ['left', 'center', 'right', 'top', 'bottom', 'dynamic']
        self.orientation = ['horizontal', 'vertical'] # the two possible orientations, the first is the default: v = vertical, h = horizontal
        
    def draw_init(self): # Default Values for Variables Set with Draw internal Arguments
        self.xside = ["lower", "upper"] # the two possibilities, the first is default
        self.yside = ["left", "right"] # the two possibilities, the first is default
        self.lines = False
        self.fill = False
        self.fill_internal = "internal"
        #self.filly = False
        self.label = None

    def bar_init(self):
        self.bar_marker = "sd"
        self.bar_fill = True # bar plot filled or not
        self.bar_width = 4 / 5 # bar width
        self.hist_bins = 10

    def confusion_matrix_init(self):
        self.cmatrix_color = 'orange+'
        self.cmatrix_style = 'bold'



# /usr/bin/env python3
# -*- coding: utf-8 -*-

# This file contains all the main plotext functions available externally to the user

##############################################
###########    Initialisation    #############
##############################################

import plotext._doc
from plotext._doc_utils import documentation as doc
from plotext._doc_utils import add

from ._figure import _figure_class
from time import sleep as _sleep
import plotext._global as _glob
from plotext import __version__
import plotext._utility as _ut

_figure = _glob.figure # the main figure at top level (defined in _global.py because it is useful also there)

##############################################
#########    Subplots Functions    ###########
##############################################

def subplots(rows = None, cols = None):
    sub = _figure._active.subplots(rows, cols)
    _figure.show() if _figure._interactive else None
    return sub

def subplot(row = None, col = None):
    sub = _figure.subplot(row, col)
    _figure.show() if _figure._interactive else None
    return sub

def main():
    return _figure.main()

def active():
    return _figure._active

##############################################
#######    Outside Set Functions    ##########
##############################################

def interactive(interactive = None):
    _figure._set_interactive(interactive)

def plot_size(width = None, height = None):
    size = _figure._active.plot_size(width, height)
    _figure.show() if _figure._interactive else None
    return size
plotsize = plot_size

def limit_size(width = None, height = None):
    #_figure._master._set_size()
    _figure._master._limit_size(width, height)
limitsize = limit_size

def take_min():
    _figure._active.take_min()
takemin = take_min

def title(label):
    _figure._active.title(label)
    _figure.show() if _figure._interactive else None

def xlabel(label = None, xside = None):
    _figure._active.xlabel(label = label, xside = xside)
    _figure.show() if _figure._interactive else None

def ylabel(label = None, yside = None):
    _figure._active.ylabel(label = label, yside = yside)
    _figure.show() if _figure._interactive else None

def xlim(left = None, right = None, xside = None):
    _figure._active.xlim(left = left, right = right, xside = xside)
    _figure.show() if _figure._interactive else None

def ylim(lower = None, upper = None, yside = None):
    _figure._active.ylim(lower = lower, upper = upper, yside = yside)
    _figure.show() if _figure._interactive else None

def xscale(scale = None, xside = None):
    _figure._active.xscale(scale = scale, xside = xside)
    _figure.show() if _figure._interactive else None

def yscale(scale = None, yside = None):
    _figure._active.yscale(scale = scale, yside = yside)
    _figure.show() if _figure._interactive else None

def xticks(ticks = None, labels = None, xside = None):
    _figure._active.xticks(ticks = ticks, labels = labels, xside = xside)
    _figure.show() if _figure._interactive else None

def yticks(ticks = None, labels = None, yside = None):
    _figure._active.yticks(ticks = ticks, labels = labels, yside = yside)
    _figure.show() if _figure._interactive else None

def xfrequency(frequency = None, xside = None):
    _figure._active.xfrequency(frequency = frequency, xside = xside)
    _figure.show() if _figure._interactive else None

def yfrequency(frequency = None, yside = None):
    _figure._active.yfrequency(frequency = frequency, yside = yside)
    _figure.show() if _figure._interactive else None

def xreverse(reverse = None, xside = None):
    _figure._active.xreverse(reverse = reverse, xside = xside)
    _figure.show() if _figure._interactive else None

def yreverse(reverse = None, yside = None):
    _figure._active.yreverse(reverse = reverse, yside = yside)
    _figure.show() if _figure._interactive else None

def xaxes(lower = None, upper = None):
    _figure._active.xaxes(lower = lower, upper = upper)
    _figure.show() if _figure._interactive else None

def yaxes(left = None, right = None):
    _figure._active.yaxes(left = left, right = right)
    _figure.show() if _figure._interactive else None

def frame(frame = None):
    _figure._active.frame(frame = frame)
    _figure.show() if _figure._interactive else None

def grid(horizontal = None, vertical = None):
    _figure._active.grid(horizontal = horizontal, vertical = vertical)
    _figure.show() if _figure._interactive else None

def canvas_color(color = None):
    _figure._active.canvas_color(color)
    _figure.show() if _figure._interactive else None

def axes_color(color = None):
    _figure._active.axes_color(color)
    _figure.show() if _figure._interactive else None

def ticks_color(color = None):
    _figure._active.ticks_color(color)
    _figure.show() if _figure._interactive else None

def ticks_style(style = None):
    _figure._active.ticks_style(style)
    _figure.show() if _figure._interactive else None

def theme(theme = None):
    _figure._active.theme(theme)
    _figure.show() if _figure._interactive else None

##############################################
###########    Clear Functions    ############
##############################################

def clear_figure(): 
    _figure._active.clear_figure()
    _figure.show() if _figure._interactive else None
clf = clear_figure

def clear_data(): 
    _figure._active.clear_data()
    _figure.show() if _figure._interactive else None
cld = clear_data

def clear_color(): 
    _figure._active.clear_color()
    _figure.show() if _figure._interactive else None
clc = clear_color

def clear_terminal(lines = None):
    _figure._active.clear_terminal(lines = lines)
clt = clear_terminal

##############################################
######    Main Plotting Functions    #########
##############################################

def scatter(*args, marker = None, color = None, style = None, fillx = None, filly = None, xside = None, yside = None, label = None):
    _figure._active.scatter(*args, xside = xside, yside = yside, marker = marker, color = color, style = style, fillx = fillx, filly = filly, label = label)
    _figure.show() if _figure._interactive else None

def plot(*args, marker = None, color = None, style = None, fillx = None, filly = None, xside = None, yside = None, label = None):
    _figure._active.plot(*args, xside = xside, yside = yside, marker = marker, color = color,  fillx = fillx, filly = filly, label = label)
    _figure.show() if _figure._interactive else None

def candlestick(dates, data, xside = None, yside = None, orientation = None, colors = None, label = None):
    _figure._active.candlestick(dates, data, xside = xside, yside = yside, orientation = orientation, colors = colors, label = label)
    _figure.show() if _figure._interactive else None

def bar(*args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, reset_ticks = None, xside = None, yside = None, label = None):
    _figure._active.bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, label = label, minimum = minimum, reset_ticks = reset_ticks)
    _figure.show() if _figure._interactive else None

def simple_bar(*args, marker = None, color = None, title = None, width = None):
    _glob.simple_bar(*args, width = width, marker = marker, color = color, title = title)
    _figure.show() if _figure._interactive else None

def multiple_bar(*args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, reset_ticks = None, xside = None, yside = None, labels = None):
    _figure._active.multiple_bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, labels = labels, minimum = minimum, reset_ticks = reset_ticks)
    _figure.show() if _figure._interactive else None

def simple_multiple_bar(*args, marker = None, colors = None, title = None, width = None, labels = None):
    _glob.simple_multiple_bar(*args, width = width, marker = marker, colors = colors, title = title, labels = labels)
    _figure.show() if _figure._interactive else None

def stacked_bar(*args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, reset_ticks = None, xside = None, yside = None, labels = None):
    _figure._active.stacked_bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, labels = labels, minimum = minimum, reset_ticks = reset_ticks)
    _figure.show() if _figure._interactive else None

def simple_stacked_bar(*args, marker = None, colors = None, title = None, width = None, labels = None):
    _glob.simple_stacked_bar(*args, width = width, marker = marker, colors = colors, title = title, labels = labels)
    _figure.show() if _figure._interactive else None

def hist(data, bins = None, marker = None, color = None, fill = None, norm = None, width = None, orientation = None, minimum = None, xside = None, yside = None, label = None):
    _figure._active.hist(data, bins = bins, norm = norm, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, label = label, minimum = minimum)
    _figure.show() if _figure._interactive else None

def candlestick(dates, data, colors = None, orientation = None, xside = None, yside = None, label = None):
    _figure._active.candlestick(dates, data, xside = xside, yside = yside, orientation = orientation, colors = colors, label = label)
    _figure.show() if _figure._interactive else None

def box(*args, quintuples = None, colors = None,  fill = None, width = None, orientation = None, minimum = None, reset_ticks = None, xside = None, yside = None, label = None):
    _figure._active.box(*args, xside = xside, yside = yside, orientation = orientation, colors = colors, label = label, fill=fill, width = width, minimum = minimum, reset_ticks = reset_ticks, quintuples = quintuples)
    _figure.show() if _figure._interactive else None

##############################################
###########    Plotting Tools    #############
##############################################

def error(*args, xerr = None, yerr = None, color = None, xside = None, yside = None, label = None):
    _figure.error(*args, xerr = xerr, yerr = yerr, xside = xside, yside = yside, color = color, label = label)
    _figure.show() if _figure._interactive else None

def event_plot(data, marker = None, color = None, orientation = None, side = None):
    _figure._active.event_plot(data, orientation = orientation, marker = marker, color = color, side = side)
    _figure.show() if _figure._interactive else None

eventplot = event_plot

def vertical_line(coordinate, color = None, xside = None):
    _figure._active.vertical_line(coordinate, color = color, xside = xside)
    _figure.show() if _figure._interactive else None
vline = vertical_line

def horizontal_line(coordinate, color = None, yside = None):
    _figure._active.horizontal_line(coordinate, color = color, yside = yside)
    _figure.show() if _figure._interactive else None
hline = horizontal_line

def text(label, x, y, color = None, background = None, style = None, orientation = None, alignment = None, xside = None, yside = None):
    _figure._active.text(label, x, y, xside = xside, yside = yside, color = color, background = background, style = style, orientation = orientation, alignment = alignment)
    _figure.show() if _figure._interactive else None

def rectangle(x = None, y = None, marker = None, color = None, lines = None, fill = None, xside = None, yside = None, label = None):
    _figure._active.rectangle(x = x, y = y, xside = xside, yside = yside, lines = lines, marker = marker, color = color, fill = fill, label = label)
    _figure.show() if _figure._interactive else None
    
def polygon(x = None, y = None,  radius = None, sides = None, marker = None, color = None, lines = None, fill = None, xside = None, yside = None, label = None):
    _figure._active.polygon(x = x, y = y, radius = radius, sides = sides, xside = xside, yside = yside, lines = lines, marker = marker, color = color, fill = fill, label = label)
    _figure.show() if _figure._interactive else None

def confusion_matrix(actual, predicted, color = None, style = None, labels = None):
    _figure._active.confusion_matrix(actual, predicted, labels = labels, color = color, style = style)
    _figure.show() if _figure._interactive else None

cmatrix = confusion_matrix

def indicator(value, label = None, color = None, style = None):
    _figure._active.indicator(value, label = label, color = color, style = style)
    _figure.show() if _figure._interactive else None

##############################################
##############    2D Plots    ################
############################################## 
    
def matrix_plot(matrix, marker = None, style = None, fast = False):
    _figure._active.matrix_plot(matrix, marker = marker, style = style, fast = fast)
    _figure.show() if _figure._interactive else None

def heatmap(dataframe, color = None, style=None):
    _figure._active.heatmap(dataframe, color = color, style = style)
    _figure.show() if _figure._interactive else None

def image_plot(path, marker = None, style = None, fast = False, grayscale = False):
    _figure._active.image_plot(path, marker = marker, style = style, grayscale = grayscale, fast = fast)
    _figure.show() if _figure._interactive else None
    
def play_gif(path):
    _glob.play_gif(path)
    _figure.show() if _figure._interactive else None

def play_video(path, from_youtube = False):
    _glob.play_video(path, from_youtube)
    _figure.show() if _figure._interactive else None

def play_youtube(url):
    _glob.play_youtube(url)
    _figure.show() if _figure._interactive else None

def get_youtube(url, path = None, log = True):
    _glob.get_youtube(url, path, log)

##############################################
##########    Build Functions    #############
##############################################

def show():
    _figure.show()

def build():
    return _figure.build()

def sleep(time = 0):
    _sleep(time)

def time(show = True):
    return _figure._get_time(show)

def save_fig(path = None, append = False, keep_colors = False):
    _figure.save_fig(path, append, keep_colors)
savefig = save_fig

def from_matplotlib(fig, marker = None):
    _glob.from_matplotlib(fig, marker = marker)

##############################################
##########     Date Functions    #############
##############################################

def date_form(input_form = None, output_form = None):
    _figure._active.date_form(input_form, output_form)
    
def set_time0(string, input_form = None):
    _figure._active.set_time0(string, input_form = input_form)

def today_datetime():
    return _figure._active.today_datetime()

def today_string(output_form = None):
    return _figure._active.today_string(output_form)

def datetime_to_string(datetime, output_form = None):
    return _figure._active.datetime_to_string(datetime, output_form = output_form)

def datetimes_to_strings(datetimes, output_form = None):
    return _figure._active.datetimes_to_strings(datetimes, output_form = output_form)

datetimes_to_string = datetimes_to_strings

def string_to_datetime(string, input_form = None):
    return _figure._active.string_to_datetime(string, input_form = input_form)

def string_to_time(string, input_form = None):##########ADD DOC############
    return _figure._active.string_to_time(string, input_form = input_form)

def strings_to_time(string, input_form = None):##########ADD DOC############
    return _figure._active.strings_to_time(string, input_form = input_form)

##############################################
##########     File Functions    ############
##############################################

script_folder = _ut.script_folder

def parent_folder(path, level = 1):
    return _ut.parent_folder(path, level = level)

def join_paths(*args):
    return _ut.join_paths(*args)

def save_text(text, path, log = True):
    _ut.save_text(text, path, log = log)

def read_data(path, delimiter = None, columns = None, first_row = None, log = True):
    return _ut.read_data(path, delimiter = delimiter, columns = columns, first_row = first_row, log = log)

def write_data(data, path, delimiter = None, columns = None, log = True):
    _ut.write_data(data, path, delimiter = delimiter, columns = columns, log = log)

def download(url, path, log = True):
    _ut.download(url, path, log)

def delete_file(path, log = True):
    _ut.delete_file(path, log = log)

test_data_url     =  _glob.test_data_url
test_bar_data_url =  _glob.test_bar_data_url
test_image_url    =  _glob.test_image_url
test_gif_url      =  _glob.test_gif_url
test_video_url    =  _glob.test_video_url
test_youtube_url  =  _glob.test_youtube_url

##############################################
##########     Other Functions    ############
##############################################

def colorize(string, color = None, style = None, background = None, show = False):
    return _ut.colorize(string, color = color, style = style, background = background, show = show)

def uncolorize(string):
    return _ut.uncolorize(string)

def terminal_size():
    return _ut.terminal_size()

ts = terminal_size

def terminal_width():
    return _ut.terminal_width()

tw = terminal_width

def terminal_height():
    return _ut.terminal_height()

th = terminal_height

def sin(periods = 2, length = 200, amplitude = 1, phase = 0, decay = 0):
    return _ut.sin(periods = periods, length = length, amplitude = amplitude, phase = phase, decay = decay)

def square(periods = 2, length = 200, amplitude = 1):
    return _ut.square(periods = periods, length = length, amplitude = amplitude)

def transpose(data):
    return _ut.transpose(data)

colors = _glob.colors

styles = _glob.styles

markers = _glob.markers

themes = _glob.themes

test = _glob.test

version = __version__

platform = _ut.platform
        
##############################################
############     Docstrings    ###############
##############################################        

add(subplots)
add(subplot)
add(main)
add(active)

add(interactive)
add(plot_size)
add(limit_size)
add(take_min)

add(title)
add(xlabel)
add(ylabel)
add(xlim)
add(ylim)
add(xscale)
add(yscale)
add(xticks)
add(yticks)
add(xfrequency)
add(yfrequency)
add(xreverse)
add(yreverse)
add(xaxes)
add(yaxes)
add(frame)
add(grid)
add(canvas_color)
add(axes_color)
add(ticks_color)
add(ticks_style)
add(theme)

add(clear_figure)
add(clear_data)
add(clear_color)
add(clear_terminal)

add(scatter)
add(plot)
add(candlestick)
add(bar)
add(simple_bar)
add(multiple_bar)
add(simple_multiple_bar)
add(stacked_bar)
add(simple_stacked_bar)
add(simple_bar)
add(hist)
add(box)

add(error)
add(event_plot)
add(vertical_line)
add(horizontal_line)
add(text)
add(rectangle)
add(polygon)
add(confusion_matrix)
add(indicator)

add(matrix_plot)
add(image_plot)

add(play_gif)
add(play_video)
add(play_youtube)
add(get_youtube)

add(show)
add(build)
add(sleep)
add(time)
add(save_fig)
add(from_matplotlib)

add(date_form)
add(set_time0)
add(today_datetime)
add(today_string)
add(datetime_to_string)
add(datetimes_to_strings)
add(string_to_datetime)

add(script_folder)
add(parent_folder)
add(join_paths)
add(save_text)
add(read_data)
add(write_data)
add(download)
add(delete_file)

add(colorize)
add(uncolorize)
add(terminal_size)
add(terminal_width)
add(terminal_height)
add(sin)
add(square)
add(transpose)

add(colors)
add(styles)
add(markers)
add(themes)
add(test)


# added by user @Freed-Wu as a replacement for package shtab if not present and in particular its method add_argument_to

FILE = None
DIRECTORY = DIR = None

def add_argument_to(parser, *args, **kwargs):
    from argparse import Action
    Action.complete = None
    return parser



from plotext._default import default_figure_class
from plotext._monitor import monitor_class
from plotext._matrix import join_matrices
from plotext._date import date_class
from plotext._doc_utils import add
from time import time as _time
import plotext._utility as ut
from time import time
import os

# A figure is a general container of either a plot (called monitor) or another figure, when subplots are nested
# This creates a hierarchy of figures, where master is the main/initial global figure, and parent is the figure containing (or above) the one considered
# The active figure is the one that can be accessed with further plotext commands - like plot(), limitsize() etc .. 
# If a figure has no sub figures, then it is used for plotting, otherwise its sub figures are checked

class _figure_class():
    
    def __init__(self, master = None, parent = None):
        self._set_family(master, parent) # it sets master, parent and active figure
        self.default = default_figure_class() # default values of figure class
        self.date = date_class()

        self._set_size(None, None) # no initial size for general figure
        self.max_or_min = max # in a matrix of subplots the maximum height/width is considered (by default) for each row/column
        
        self.monitor = monitor_class() if self._is_master else self._parent.monitor.copy() # each figure has a monitor for plotting; which by default is deep copied from its parent figure monitor, so that a figure with multiple sub plots can have easily the same plot and plot settings and preferences (without rewriting code)
        self.monitor.set_date(self.date) # to make sure that the date settings of a figure are the same for its subplots

        self._set_master() if self._is_master else None # it sets the master figure size and other utilities 

        self._set_slots_max(*self._master._size) # sets the maximum number of sub figures in the current figure (from master figure size)
        self.subplots(0, 0) # no sub figures added by default, so that the current figure is used for plotting
        
    def _set_family(self, master = None, parent = None):
        self._parent = self if parent is None else parent # the figure just above this one
        self._master = self if master is None else master # the figure above all others
        self._is_master = self is self._master
        self._active = self if self._is_master else self._master._active # the active figure, such that further plotting or settings commands refer to it

    def _set_master(self):
        self._limit_size(True, True)  # limit size (only available for master, as sub figures are by default limited by parent figure)
        self._set_terminal_size(*ut.terminal_size()) # get and set terminal size
        self._set_master_size()  # set master size to terminal
        self._time = None # computational time of show() method, only available for global figure
        self._dummy = _figure_class(self._master, self._master) # the master has a dumm container for subplots that do not actually exist (anymore due to change of size)
        self._master.monitor.set_size(self._size)
        self._dummy.monitor.set_size(self._size)
        self._set_interactive() # if to make the final figure interactive: every command gets directly printed

##############################################
###########    Size Functions    #############
##############################################

    def _set_interactive(self, interactive = None):
        self._interactive = self.default.interactive if interactive is None else bool(interactive)

    def _set_size(self, width = None, height = None):
        self._width = None if width is None else int(width)
        self._height = None if height is None else int(height)
        self._size = [self._width, self._height]

    def _limit_size(self, width = None, height = None):
        self._limit_width = self.default.limit_width if width is None else bool(width)
        self._limit_height = self.default.limit_height if height is None else bool(height)
        self._limit = [self._limit_width, self._limit_height]

    def _set_terminal_size(self, width = None, height = None):
        self._width_term = self.default._width_term if width is None else width
        extra_lines = 2 if ut.is_ipython() else 1
        self._height_term = self.default._height_term if height is None else max(height - extra_lines, 0)
        self._size_term = [self._width_term, self._height_term]

    def _set_master_size(self):
        width = self._width_term if self._width is None or (self._width > self._width_term and self._limit_width) else self._width
        height = self._height_term if self._height is None or (self._height > self._height_term and self._limit_height) else self._height
        self._set_size(width, height)
        
##############################################
#########    Subplots Functions    ###########
##############################################

    def _set_slots_max(self, width = None, height = None):
        self._rows_max = height # (height + 1) // 3 
        self._cols_max = width # (width + 1) // 3 
        self._slots_max = [self._rows_max, self._cols_max]

    def _set_slots(self, rows = None, cols = None):
        rows = 1 if rows is None else int(abs(rows))
        cols = 1 if cols is None else int(abs(cols))
        self._rows = min(rows, self._rows_max)
        self._cols = min(cols, self._cols_max)
        self._Rows = list(range(1, self._rows + 1))
        self._Cols = list(range(1, self._cols + 1))
        self._slots = [self._rows, self._cols]
        self._no_plots = 0 in self._slots #or self._is_master
        
    def _set_subplots(self):
        self.subfig = [[_figure_class(self._master, self) for col in self._Cols] for row in self._Rows]
        
    def _get_subplot(self, row = None, col = None):
        return self.subfig[row - 1][col - 1] if row in self._Rows and col in self._Cols else self._master._dummy

    def subplot(self, row = None, col = None):
        row = 1 if row is None else int(abs(row))
        col = 1 if col is None else int(abs(col))
        active = self._get_subplot(row, col)
        self._active = active
        self._master._active = active
        return self._master._active

    def subplots(self, rows = None, cols = None):
        self._set_slots(rows, cols)
        self._set_subplots()
        return self

##############################################
#######    External Set Functions    #########
##############################################

    def title(self, label = None):
        self.monitor.set_title(label) if self._no_plots else [[self._get_subplot(row, col).title(label) for col in self._Cols] for row in self._Rows]

    def xlabel(self, label = None, xside = None):
        self.monitor.set_xlabel(label = label, xside = xside) if self._no_plots else [[self._get_subplot(row, col).xlabel(label = label, xside = xside) for col in self._Cols] for row in self._Rows]
        
    def ylabel(self, label = None, yside = None):
        self.monitor.set_ylabel(label = label, yside = yside) if self._no_plots else [[self._get_subplot(row, col).ylabel(label = label, yside = yside) for col in self._Cols] for row in self._Rows]

    def xlim(self, left = None, right = None, xside = None):
        self.monitor.set_xlim(left = left, right = right, xside = xside) if self._no_plots else [[self._get_subplot(row, col).xlim(left = left, right = right, xside = xside) for col in self._Cols] for row in self._Rows]

    def ylim(self, lower = None, upper = None, yside = None):
        self.monitor.set_ylim(lower = lower, upper = upper, yside = yside) if self._no_plots else [[self._get_subplot(row, col).ylim(lower = lower, upper = upper, yside = yside) for col in self._Cols] for row in self._Rows]
        
    def xscale(self, scale = None, xside = None):
        self.monitor.set_xscale(scale = scale, xside = xside) if self._no_plots else [[self._get_subplot(row, col).xscale(scale = scale, xside = xside) for col in self._Cols] for row in self._Rows]
        
    def yscale(self, scale = None, yside = None):
        self.monitor.set_yscale(scale = scale, yside = yside) if self._no_plots else [[self._get_subplot(row, col).yscale(scale = scale, yside = yside) for col in self._Cols] for row in self._Rows]
        
    def xticks(self, ticks = None, labels = None, xside = None):
        self.monitor.set_xticks(ticks = ticks, labels = labels, xside = xside) if self._no_plots else [[self._get_subplot(row, col).xticks(ticks = ticks, labels = labels, xside = xside) for col in self._Cols] for row in self._Rows]

    def yticks(self, ticks = None, labels = None, yside = None):
        self.monitor.set_yticks(ticks = ticks, labels = labels, yside = yside) if self._no_plots else [[self._get_subplot(row, col).yticks(ticks = ticks, labels = labels, yside = yside) for col in self._Cols] for row in self._Rows]

    def xfrequency(self, frequency = None, xside = None):
        self.monitor.set_xfrequency(frequency = frequency, xside = xside) if self._no_plots else [[self._get_subplot(row, col).xfrequency(frequency = frequency, xside = xside) for col in self._Cols] for row in self._Rows]

    def yfrequency(self, frequency = None, yside = None):
        self.monitor.set_yfrequency(frequency = frequency, yside = yside) if self._no_plots else [[self._get_subplot(row, col).yfrequency(frequency = frequency, yside = yside) for col in self._Cols] for row in self._Rows]

    def xreverse(self, reverse = None, xside = None):
        self.monitor.set_xreverse(reverse = reverse, xside = xside) if self._no_plots else [[self._get_subplot(row, col).xreverse(reverse = reverse, xside = xside) for col in self._Cols] for row in self._Rows]

    def yreverse(self, reverse = None, yside = None):
        self.monitor.set_yreverse(reverse = reverse, yside = yside) if self._no_plots else [[self._get_subplot(row, col).yreverse(reverse = reverse, yside = yside) for col in self._Cols] for row in self._Rows]

    def xaxes(self, lower = None, upper = None):
        self.monitor.set_xaxes(lower = lower, upper = upper) if self._no_plots else [[self._get_subplot(row, col).xaxes(lower = lower, upper = upper) for col in self._Cols] for row in self._Rows]

    def yaxes(self, left = None, right = None):
        self.monitor.set_yaxes(left = left, right = right) if self._no_plots else [[self._get_subplot(row, col).yaxes(left = left, right = right) for col in self._Cols] for row in self._Rows]

    def frame(self, frame = None):
        self.monitor.set_frame(frame = frame) if self._no_plots else [[self._get_subplot(row, col).frame(frame = frame) for col in self._Cols] for row in self._Rows]
        
    def grid(self, horizontal = None, vertical = None):
        self.monitor.set_grid(horizontal = horizontal, vertical = vertical) if self._no_plots else [[self._get_subplot(row, col).grid(horizontal = horizontal, vertical = vertical) for col in self._Cols] for row in self._Rows]

    def canvas_color(self, color = None):
        self.monitor.set_canvas_color(color) if self._no_plots else [[self._get_subplot(row, col).canvas_color(color) for col in self._Cols] for row in self._Rows]

    def axes_color(self, color = None):
        self.monitor.set_axes_color(color) if self._no_plots else [[self._get_subplot(row, col).axes_color(color) for col in self._Cols] for row in self._Rows]

    def ticks_color(self, color = None):
        self.monitor.set_ticks_color(color) if self._no_plots else [[self._get_subplot(row, col).ticks_color(color) for col in self._Cols] for row in self._Rows]
        
    def ticks_style(self, style = None):
        self.monitor.set_ticks_style(style) if self._no_plots else [[self._get_subplot(row, col).ticks_style(style) for col in self._Cols] for row in self._Rows]
        
    def theme(self, theme = None):
        self.monitor.set_theme(theme) if self._no_plots else [[self._get_subplot(row, col).theme(theme) for col in self._Cols] for row in self._Rows]

##############################################
###########    Clear Functions    ############
###########################x##################

    def clear_figure(self):
        self.__init__()# if self._no_plots else [[self._get_subplot(row, col).clear_figure() for col in self._Cols] for row in self._Rows]
    clf = clear_figure
    
    def clear_data(self):
        self.monitor.data_init() if self._no_plots else [[self._get_subplot(row, col).clear_data() for col in self._Cols] for row in self._Rows]
    cld = clear_data

    def clear_color(self):
        self.monitor.clear_color() if self._no_plots else [[self._get_subplot(row, col).clear_color() for col in self._Cols] for row in self._Rows]
    clc = clear_color

    def clear_terminal(self, lines = None):
        ut.clear_terminal(lines = lines)
    clt = clear_terminal

##############################################
###########    Plot Functions    #############
##############################################

    def _draw(self, *args, **kwargs):
        self.monitor.draw(*args, **kwargs) if self._no_plots else [[self._get_subplot(row, col)._draw(*args, **kwargs) for col in self._Cols] for row in self._Rows]

    def scatter(self, *args, marker = None, color = None, style = None, fillx = None, filly = None, xside = None, yside = None, label = None):
        self._draw(*args, xside = xside, yside = yside, lines = False, marker = marker, color = color, style = style, fillx = fillx, filly = filly, label = label)

    def plot(self, *args, marker = None, color = None, style = None, fillx = None, filly = None, xside = None, yside = None, label = None):
        self._draw(*args, xside = xside, yside = yside, lines = True, marker = marker, color = color,  fillx = fillx, filly = filly, label = label)
        
    def bar(self, *args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, reset_ticks = None, xside = None, yside = None, label = None):
        self.monitor.draw_bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, label = label, minimum = minimum, reset_ticks = reset_ticks) if self._no_plots else [[self._get_subplot(row, col).bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, label = label, minimum = minimum, reset_ticks = reset_ticks) for col in self._Cols] for row in self._Rows]

    def multiple_bar(self, *args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, reset_ticks = None, xside = None, yside = None, labels = None):
        self.monitor.draw_multiple_bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, labels = labels, minimum = minimum, reset_ticks = reset_ticks) if self._no_plots else [[self._get_subplot(row, col).multiple_bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, label = label, minimum = minimum, reset_ticks = reset_ticks) for col in self._Cols] for row in self._Rows]
    
    def stacked_bar(self, *args, marker = None, color = None, fill = None, width = None, orientation = None, minimum = None, reset_ticks = None, xside = None, yside = None, labels = None):
        self.monitor.draw_stacked_bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, labels = labels, minimum = minimum, reset_ticks = reset_ticks) if self._no_plots else [[self._get_subplot(row, col).stacked_bar(*args, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, label = label, minimum = minimum, reset_ticks = reset_ticks) for col in self._Cols] for row in self._Rows]

    def hist(self, data, bins = None, marker = None, color = None, fill = None, norm = None, width = None, orientation = None, minimum = None, xside = None, yside = None, label = None):
        self.monitor.draw_hist(data, bins = bins, norm = norm, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, label = label, minimum = minimum) if self._no_plots else [[self._get_subplot(row, col).hist(data, bins = bins, norm = norm, xside = xside, yside = yside, marker = marker, color = color, fill = fill, width = width, orientation = orientation, label = label, minimum = minimum) for col in self._Cols] for row in self._Rows]

    def candlestick(self, dates, data, colors = None, orientation = None, xside = None, yside = None, label = None):
        self.monitor.draw_candlestick(dates, data, xside = xside, yside = yside, orientation = orientation, colors = colors, label = label) if self._no_plots else [[self._get_subplot(row, col).candlestick(dates, data, orientation = orientation, colors = colors, label = label) for col in self._Cols] for row in self._Rows]

    def box(self, *args, quintuples = None, colors = None,  fill = None, width = None, orientation = None, minimum = None, reset_ticks = None, xside = None, yside = None, label = None):
        self.monitor.draw_box(*args, xside = xside, yside = yside, orientation = orientation, colors = colors, label = label, fill = fill, width = width, minimum = minimum, reset_ticks = reset_ticks, quintuples = quintuples) if self._no_plots else [[self._get_subplot(row, col).box(*args, orientation = orientation, colors = colors, label = label, fill = fill, width = width, minimum = minimum, reset_ticks = reset_ticks, quintuples = quintuples) for col in self._Cols] for row in self._Rows]

##############################################
###########    Plotting Tools    #############
##############################################

    def error(self, *args, xerr = None, yerr = None, color = None, xside = None, yside = None, label = None):
        self.monitor.draw_error(*args, xerr = xerr, yerr = yerr, xside = xside, yside = yside, color = color, label = label) if self._no_plots else [[self._get_subplot(row, col).error(*args, xerr = xerr, yerr = yerr, xside = xside, yside = yside, color = color, label = label) for col in self._Cols] for row in self._Rows]

    def event_plot(self, data, marker = None, color = None, orientation = None, side = None):
        self.monitor.draw_event_plot(data, orientation = orientation, marker = marker, color = color, side = side) if self._no_plots else [[self._get_subplot(row, col).event_plot(data, orientation = orientation, marker = marker, color = color, side = side) for col in self._Cols] for row in self._Rows]
    eventplot = event_plot

    def vertical_line(self, coordinate, color = None, xside = None):
        self.monitor.draw_vertical_line(coordinate, color = color, xside = xside) if self._no_plots else [[self._get_subplot(row, col).vertical_line(coordinate, color = color, xside = xside) for col in self._Cols] for row in self._Rows]
    vline = vertical_line
        
    def horizontal_line(self, coordinate, color = None, yside = None):
        self.monitor.draw_horizontal_line(coordinate, color = color, yside = yside) if self._no_plots else [[self._get_subplot(row, col).horizontal_line(coordinate, color = color, yside = yside) for col in self._Cols] for row in self._Rows]     
    hline = horizontal_line

    def text(self, label, x, y, color = None, background = None, style = None, orientation = None, alignment = None, xside = None, yside = None):
        self.monitor.draw_text(label, x, y, xside = xside, yside = yside, color = color, background = background, style = style, orientation = orientation, alignment = alignment) if self._no_plots else [[self._get_subplot(row, col).text(label, x, y, xside = xside, yside = yside, color = color, background = background, style = style, orientation = orientation, alignment = alignment) for col in self._Cols] for row in self._Rows]

    def rectangle(self, x = None, y = None, marker = None, color = None, lines = None, fill = None, xside = None, yside = None, label = None):
        self.monitor.draw_rectangle(x = x, y = y, xside = xside, yside = yside, lines = lines, marker = marker, color = color, fill = fill, label = label) if self._no_plots else [[self._get_subplot(row, col).rectangle(x = x, y = y, xside = xside, yside = yside, lines = lines, marker = marker, color = color, fill = fill, label = label) for col in self._Cols] for row in self._Rows]

    def polygon(self, x = None, y = None,  radius = None, sides = None, marker = None, color = None, lines = None, fill = None, xside = None, yside = None, label = None):
        self.monitor.draw_polygon(x = x, y = y, radius = radius, sides = sides, xside = xside, yside = yside, lines = lines, marker = marker, color = color, fill = fill, label = label) if self._no_plots else [[self._get_subplot(row, col).polygon(x = x, y = y, radius = radius, sides = sides, xside = xside, yside = yside, lines = lines, marker = marker, color = color, fill = fill, label = label) for col in self._Cols] for row in self._Rows]

    def confusion_matrix(self, actual, predicted, color = None, style = None, labels = None):
        self.monitor.draw_confusion_matrix(actual, predicted, labels = labels, color = color, style = style) if self._no_plots else [[self._get_subplot(row, col).confusion_matrix(actual, predicted, labels = labels, color = color, style = style) for col in self._Cols] for row in self._Rows]

    cmatrix = confusion_matrix

    def indicator(self, value, label = None, color = None, style = None):
        self.monitor.draw_indicator(value, label = label, color = color, style = style) if self._no_plots else [[self._get_subplot(row, col).confusion_matrix(value, label = label, color = color, style = style) for col in self._Cols] for row in self._Rows]

##############################################
##############    2D Plots    ################
############################################## 

    def matrix_plot(self, matrix, marker = None, style = None, fast = False):
        self.monitor.draw_matrix(matrix, marker = marker, style = style, fast = fast) if self._no_plots else [[self._get_subplot(row, col).matrix_plot(matrix, marker = marker, style = style, fast = fast) for col in self._Cols] for row in self._Rows]

    def heatmap(self, dataframe, color = None, style = None):
        self.monitor.draw_heatmap(dataframe, color = color, style = style) if self._no_plots else [[self._get_subplot(row, col).heatmap(dataframe, color = color, style = style) for col in self._Cols] for row in self._Rows]

    def image_plot(self, path, marker = None, style = None, fast = False, grayscale = False):
        self.monitor.draw_image(path, marker = marker, style = style, grayscale = grayscale, fast = fast) if self._no_plots else [[self._get_subplot(row, col).image_plot(path, marker = marker, style = style, grayscale = grayscale, fast = fast) for col in self._Cols] for row in self._Rows]
        
##############################################
###########    Date Functions    #############
##############################################

    def date_form(self, input_form = None, output_form = None):
        self._master._dummy.date.date_form(input_form, output_form)
        if self._no_plots:
            self.monitor.date.date_form(input_form, output_form)
        else:
            [[self._get_subplot(row, col).date_form(input_form, output_form) for col in self._Cols] for row in self._Rows]
        
    def set_time0(self, string, input_form = None):
        self.monitor.date.set_time0(string, form) if self._no_plots else [[self._get_subplot(row, col).set_time0(string, form) for col in self._Cols] for row in self._Rows]

    def today_datetime(self):
        return self.monitor.date.today_datetime()
    
    def today_string(self, output_form = None):
        return self.monitor.date.today_string(output_form)
    
    def datetime_to_string(self, datetime, output_form = None):
        return self.monitor.date.datetime_to_string(datetime, output_form = output_form)
    
    def datetimes_to_strings(self, datetimes, output_form = None):
        return self.monitor.date.datetimes_to_strings(datetimes, output_form = output_form)
        
    def string_to_datetime(self, string, input_form = None):
        return self.monitor.date.string_to_datetime(string, input_form = input_form)
    
    def string_to_time(self, string, input_form = None):
        return self.monitor.date.string_to_time(string, input_form = input_form)
    
    def strings_to_time(self, string, input_form = None):
        return self.monitor.date.strings_to_time(string, input_form = input_form)

##############################################
###########    Build Functions    ############
##############################################

    def show(self): # it build and shows the overall figure
        t = time()
        self.build() 
        ut.write(self.monitor.matrix.canvas) # it prints the final canvas
        self._time = time() - t # computational time of build + print (it does not include any pre-processing time, which is gets more important for bar and image plots)
        self.main() if not self._master._interactive else None# it returns control to main figure on top level

    def build(self): # it build the current figure without showing it
        self._set_sizes()
        self._build_matrix()
        self.monitor.matrix.set_canvas() if not self.monitor.fast_plot else None
        return self.monitor.matrix.get_canvas() 

    def _build_matrix(self):
        if self._no_plots:
            self.monitor.build_plot() if not self.monitor.fast_plot else None
        else:
            [[self._get_subplot(row, col)._build_matrix() for col in self._Cols] for row in self._Rows]
            matrices = [[self._get_subplot(row, col).monitor.matrix for col in self._Cols] for row in self._Rows]
            self.monitor.matrix = join_matrices(matrices)

##############################################
#########    Set Size Utilities    ###########
##############################################

    def _set_sizes(self): # it properly sets coherent sub figure dimensions
        self._set_slots_max(*self._size) 
        self._set_slots(*self._slots)
        widths = self._get_widths()
        widths = ut.set_sizes(widths, self._width) # it sets the free subplots widths in accord with the parent figure width
        widths = ut.fit_sizes(widths, self._width) # it fits the subplots widths to the parent figure width 
        heights = self._get_heights()
        heights = ut.set_sizes(heights, self._height) # it sets the free subplots height in accord with the parent figure height 
        heights = ut.fit_sizes(heights, self._height) # it fits the subplots heights to the parent figure height
        width = sum(widths) if len(widths) > 1 else self._width
        height = sum(heights) if len(widths) > 1 else self._height

        self.monitor.set_size(self._size)
        # self._set_size(width, height)
        [[self._set_subplot_size(row, col, widths[col - 1], heights[row - 1]) for col in self._Cols] for row in self._Rows] if (not self._no_plots) else None # to make sure that all sub figures have internal dimensions set as well 

    def _get_widths(self): # the subplots max/min widths for each column
        widths = [[self._get_subplot(row, col)._width for row in self._Rows] for col in self._Cols]
        widths = [self.max_or_min([sub for sub in el if sub is not None], default = None) for el in widths]
        return widths

    def _get_heights(self): # the subplots max/min heights for each row
        heights = [[self._get_subplot(row, col)._height for col in self._Cols] for row in self._Rows]
        heights = [self.max_or_min([sub for sub in el if sub is not None], default = None) for el in heights]
        return heights

    def _set_subplot_size(self, row = None, col = None, width = None, height = None):
        self._get_subplot(row, col)._set_size(width, height)
        self._get_subplot(row, col)._set_sizes() 

##############################################
######    Externally Called Utility    #######
##############################################

    def _get_time(self, show = True): # it returns the computational time of latest show or build function
        time = ut.format_time(self._time)
        print(ut.format_strings("plotext time:", time, ut.title_color)) if show else None
        return self._time

    def main(self): # returns the master figure and sets the active figure to the master
        self._master._active = self._master
        return self._master

    def plot_size(self, width = None, height = None):
        width = self._width if width is None else width
        height = self._height if height is None else height
        self._set_size(width, height)
        self._set_master_size() if self._is_master else None
        self.monitor.size = self._size
        return self._width, self._height
        
    plotsize = plot_size

    def take_min(self): # in a matrix of subplots the maximum height/width will be considered for each row/column
        self.max_or_min = min

    def save_fig(self, path = None, append = False, keep_colors = False): # it saves the plot as text or html, keep_colors = True preserves ansi colors for texts
        path = 'plotext.txt' if path is None or not ut.correct_path(path) else path
        _, extension = os.path.splitext(path)
        canvas = self.monitor.matrix.get_canvas()
        if extension == ".html":
            canvas = self.monitor.matrix.to_html()
        elif not keep_colors:
            canvas = ut.uncolorize(canvas)
        ut.save_text(canvas, path, append)
    savefig = save_fig
    
##############################################
############     Docstrings    ###############
##############################################
    add(subplots) 
    add(subplots)
    add(subplot)
    add(main)
    
    add(plot_size)
    add(take_min)
    
    add(title)
    add(xlabel)
    add(ylabel)
    add(xlim)
    add(ylim)
    add(xscale)
    add(yscale)
    add(xticks)
    add(yticks)
    add(xfrequency)
    add(yfrequency)
    add(xreverse)
    add(yreverse)
    add(xaxes)
    add(yaxes)
    add(frame)
    add(grid)
    add(canvas_color)
    add(axes_color)
    add(ticks_color)
    add(ticks_style)
    add(theme)
    
    add(clear_figure)
    add(clear_data)
    add(clear_color)
    add(clear_terminal)
    
    add(scatter)
    add(plot)
    add(bar)
    add(multiple_bar)
    add(stacked_bar)
    add(hist)
    add(candlestick)
    add(box)

    add(error)
    add(event_plot)
    add(vertical_line)
    add(horizontal_line)
    add(text)
    add(rectangle)
    add(polygon)
    add(confusion_matrix)
    add(indicator)

    add(matrix_plot)
    add(image_plot)
    
    add(show)
    add(build)
    add(save_fig)
    
    add(date_form)
    add(set_time0)
    add(today_datetime)
    add(today_string)
    add(datetime_to_string)
    add(datetimes_to_strings)
    add(string_to_datetime)


import sys, shutil, os, re, math, inspect
from plotext._dict import *

###############################################
#########    Number Manipulation     ##########
###############################################

def round(n, d = 0): # the standard round(0.5) = 0 instead of 1; this version rounds 0.5 to 1
    n *= 10 ** d
    f = math.floor(n)
    r = f if n - f < 0.5 else math.ceil(n)
    return r * 10 ** (-d)

def mean(x, y, p = 1): # mean of x and y with optional power p; if p tends to 0 the minimum is returned; if p tends to infinity the max is returned; p = 1 is the standard mean
    return ((x ** p + y ** p) / 2) ** (1 / p)

def replace(data, data2, element = None): # replace element in data with correspondent in data2 when element is found
    res = []
    for i in range(len(data)):
        el = data[i] if data[i] != element else data2[i]
        res.append(el)
    return res

def try_float(data): # it turn a string into float if it can
    try:
        return float(data)
    except:
        return data

def quantile(data, q): # calculate the quantile of a given array
    data = sorted(data)
    index = q * (len(data) - 1)
    if index.is_integer():
        return data[int(index)]
    else:
        return (data[int(index)] + data[int(index) + 1]) / 2

###############################################
###########    List Creation     ##############
###############################################

def linspace(lower, upper, length = 10): # it returns a lists of numbers from lower to upper with given length
    slope = (upper - lower) / (length - 1) if length > 1 else 0
    return [lower + x * slope for x in range(length)]

def sin(periods = 2, length = 200, amplitude = 1, phase = 0, decay = 0): # sinusoidal data with given parameters
    f = 2 * math.pi * periods / (length - 1)
    phase =  math.pi * phase
    d = decay / length
    return [amplitude * math.sin(f * el + phase) * math.exp(- d * el) for el in range(length)]

def square(periods = 2, length = 200, amplitude = 1):
    T = length / periods
    step = lambda t: amplitude if t % T <= T / 2 else - amplitude
    return [step(i) for i in range(length)]

def to_list(data, length): # eg: to_list(1, 3) = [1, 1 ,1]; to_list([1,2,3], 6) = [1, 2, 3, 1, 2, 3]
    data = data if isinstance(data, list) else [data] * length
    data = data * math.ceil(length / len(data)) if len(data) > 0 else []
    return data[ : length]

def difference(data1, data2) : # elements in data1 not in date2
    return [el for el in data1 if el not in data2]

###############################################
#########    List Transformation     ##########
###############################################

def log(data): # it apply log function to the data
    return [math.log10(el) for el in data] if isinstance(data, list) else math.log10(data)

def power10(data): # it apply log function to the data
    return [10 ** el for el in data]

def floor(data): # it floors a list of data
    return list(map(math.floor, data))

def repeat(data, length): # repeat the same data till length is reached
    l = len(data) if type(data) == list else 1
    data = join([data] * math.ceil(length / l))
    return data[ : length]

###############################################
##########    List Manipulation     ###########
###############################################

def no_duplicates(data): # removes duplicates from a list
    return list(set(list(data)))
    #return list(dict.fromkeys(data)) # it takes double time

def join(data): # flatten lists at first level
    #return [el for row in data for el in row]
    return [el for row in data for el in (join(row) if type (row) == list else [row])]

def cumsum(data): # it returns the cumulative sums of a list; eg: cumsum([0,1,2,3,4]) = [0,1,3,6,10]
    s = [0]
    for i in range(len(data)):
        s.append(s[-1] + data[i])
    return s[1:]

###############################################
#########    Matrix Manipulation     ##########
###############################################

def matrix_size(matrix):  # cols, height
    return [len(matrix[0]), len(matrix)] if matrix != [] else [0, 0]

def transpose(data, length = 1): # it needs no explanation
    return [[]] * length if data == [] else list(map(list, zip(*data)))

def vstack(matrix, extra): # vertical stack of two matrices
    return extra + matrix # + extra

def hstack(matrix, extra): # horizontal stack of two matrices
    lm, le = len(matrix), len(extra)
    l = max(lm, le)
    return [matrix[i] + extra[i] for i in range(l)]

def turn_gray(matrix): # it takes a standard matrix and turns it into an grayscale one
    M, m = max(join(matrix), default = 0), min(join(matrix), default = 0)
    to_gray = lambda el: tuple([int(255 * (el - m) / (M - m))] * 3) if M != m else (127, 127, 127)
    return [[to_gray(el) for el in l] for l in matrix]

def brush(*lists): # remove duplicates from lists x, y, z ...
    l = min(map(len, lists))
    lists = [el[:l] for el in lists]
    z = list(zip(*lists))
    z = no_duplicates(z)
    #z = sorted(z)#, key = lambda x: x[0])
    lists = transpose(z, len(lists))
    return lists

###############################################
#########   String Manipulation     ###########
###############################################

nl = "\n"

def only_spaces(string): # it returns True if string is made of only empty spaces or is None or ''
    return (type(string) == str) and (string == len(string) * space) #and len(string) != 0

def format_time(time): # it properly formats the computational time
    t = time if time is not None else 0
    unit = 's' if t >= 1 else 'ms' if t >= 10 ** -3 else 'µs'
    p = 0 if unit == 's' else 3 if unit == 'ms' else 6
    t = round(10 ** p * t, 1)
    l = len(str(int(t)))
    t = str(t)
    #t = ' ' * (3 - l) + t
    return t[ : l + 2] + ' ' + unit

positive_color = 'green+'
negative_color = 'red'
title_color = 'cyan+'

def format_strings(string1, string2, color = positive_color): # returns string1 in bold and with color + string2 with a pre-formatted style
    return colorize(string1, color, "bold") + " " + colorize(string2, style = info_style)

def correct_coord(string, label, coord): # In the attempt to insert a label in string at given coordinate, the coordinate is adjusted so not to hit the borders of the string
    l = len(label)
    b, e = max(coord - l + 1, 0), min(coord + l, len(string) - 1)
    data = [i for i in range(b, e) if string[i] is space]
    b, e = min(data, default = coord - l + 1), max(data, default = coord + l)
    b, e = e - l + 1, b + l
    return (b + e - l) // 2

def no_char_duplicates(string, char): # it remove char duplicates from string
    pattern = char + '{2,}'
    string = re.sub(pattern, char, string)
    return string

def read_lines(text, delimiter = None, columns = None): # from a long text to well formatted data
    delimiter = " " if delimiter is None else delimiter
    data = []
    columns = len(no_char_duplicates(text[0], delimiter).split(delimiter)) if columns is None else columns
    for i in range(len(text)):
        row = text[i]
        row = no_char_duplicates(row, delimiter)
        row = row.split(delimiter)
        row = [el.replace('\n', '') for el in row]
        cols = len(row)
        row = [row[col].replace('\n', '') if col in range(cols) else '' for col in range(columns)]
        row = [try_float(el) for el in row]
        data.append(row)
    return data

def pad_string(num, length): # pad a number with spaces before to reach length
    num = str(num)
    l = len(num)
    return num + ' ' * (length - l)

def max_length(strings):
    strings = map(str, strings)
    return max(map(len, strings), default = 0)

###############################################
##########   File Manipulation     ############
###############################################

def correct_path(path):
    folder, base = os.path.dirname(path), os.path.basename(path)
    folder = os.path.expanduser("~") if folder in ['', '~'] else folder
    path = os.path.join(folder, base)
    return path

def is_file(path, log = True): # returns True if path exists
    res = os.path.isfile(path)
    print(format_strings("not a file:", path, negative_color)) if not res and log else None
    return res

def script_folder(): # the folder of the script executed
    return parent_folder(inspect.getfile(sys._getframe(1)))

def parent_folder(path, level = 1): # it return the parent folder of the path or file given; if level is higher then 1 the process is iterated
    if level <= 0:
        return path
    elif level == 1:
        return os.path.abspath(os.path.join(path, os.pardir))
    else:
        return parent_folder(parent_folder(path, level - 1))

def join_paths(*args): # it join a list of string in a proper file path; if the first argument is ~ it is turnded into the used home folder path
    args = list(args)
    args[0] = _correct_path(args[0]) if args[0] == "~" else args[0]
    return os.path.abspath(os.path.join(*args))

def delete_file(path, log = True): # remove the file if it exists
    path = correct_path(path)
    if is_file(path):
        os.remove(path)
        print(format_strings("file removed:", path, negative_color)) if log else None

def read_data(path, delimiter = None, columns = None, first_row = None, log = True): # it turns a text file into data lists
    path = correct_path(path)
    first_row = 0 if first_row is None else int(first_row)
    file = open(path, "r")
    text = file.readlines()[first_row:]
    file.close()
    print(format_strings("data read from", path)) if log else None
    return read_lines(text, delimiter, columns)

def write_data(data, path, delimiter = None, columns = None, log = True): # it turns a matrix into a text file
    delimiter = " " if delimiter is None else delimiter
    cols = len(data[0])
    cols = range(1, cols + 1) if columns is None else columns
    text = ""
    for row in data:
        row = [row[i - 1] for i in cols]
        row = list(map(str, row))
        text += delimiter.join(row) + '\n'
    save_text(text, path, log = log)

def save_text(text, path, append = False, log = True): # it saves some text to the path selected
    path = correct_path(path)
    mode = "a" if append else "w+"
    with open(path , mode, encoding='utf-8') as file:
        file.write(text)
    print(format_strings("text saved in", path)) if log else None

def download(url, path, log = True): # it download the url (image, video, gif etc) to path
    from urllib.request import urlretrieve
    path = correct_path(path)
    urlretrieve(url, path)
    print(format_strings('url saved in', path)) if log else None

###############################################
#########    Platform Utilities    ############
###############################################

def is_ipython(): # true if running in ipython shenn
    try:
        __IPYTHON__
        return True
    except NameError:
        return False

def platform(): # the platform (unix or windows) you are using plotext in
   platform = sys.platform
   if platform in {'win32', 'cygwin'}:
       return 'windows'
   else:
       return 'unix'

platform = platform()

# to enable ascii escape color sequences
if platform == "windows":
    import subprocess
    subprocess.call('', shell = True)

def terminal_size(): # it returns the terminal size as [width, height]
    try:
        size = shutil.get_terminal_size()
        return list(size)
    except OSError:
        return [None, None]

terminal_width = lambda: terminal_size()[0]
tw = terminal_width

terminal_height = lambda: terminal_size()[1]
th = terminal_height

def clear_terminal(lines = None): # it cleat the entire terminal, or the specified number of lines
    if lines is None:
        write('\033c')
    else:
        for r in range(lines):
            write("\033[A") # moves the curson up
            write("\033[2K") # clear the entire line

def write(string): # the print function used by plotext
    sys.stdout.write(string)

class memorize: # it memorise the arguments of a function, when used as its decorator, to reduce computational time
    def __init__(self, f):
        self.f = f
        self.memo = {}
    def __call__(self, *args):
        if not args in self.memo:
            self.memo[args] = self.f(*args)
        return self.memo[args]

##############################################
#########    Marker Utilities      ###########
##############################################

space = ' ' # the default null character that appears as background to all plots
plot_marker = "hd" if platform == 'unix' else 'dot'

hd_markers = {hd_codes[el] : el for el in hd_codes}
fhd_markers = {fhd_codes[el] : el for el in fhd_codes}
braille_markers = {braille_codes[el] : el for el in braille_codes}
simple_bar_marker = '▇'

@memorize
def get_hd_marker(code):
   return hd_codes[code] if len(code) == 4 else fhd_codes[code] if len(code) == 6 else braille_codes[code]

def marker_factor(marker, hd, fhd, braille): # useful to improve the resolution of the canvas for higher resolution markers
   return hd if marker == 'hd' else fhd if marker == 'fhd' else braille if marker == 'braille' else 1

##############################################
###########    Color Utilities    ############
##############################################

# A user could specify three types of colors
  # an integer for 256 color codes
  # a tuple    for RGB color codes
  # a string   for 16 color codes or styles

# Along side the user needs to specify whatever it is for background / fullground / style
# which plotext calls 'character' = 0 / 1 / 2


#colors_no_plus = [el for el in colors if '+' not in el and el + '+' not in colors and el is not no_color] # basically just [black, white]

def get_color_code(color): # the color number code from color string
    color = color.strip()
    return color_codes[color]

def get_color_name(code): # the color string from color number code
    codes = list(color_codes.values())
    return colors[codes.index(code)] if code in codes else no_color

def is_string_color(color):
    return isinstance(color, str) and color.strip() in colors

def is_integer_color(color):
    return isinstance(color, int) and 0 <= color <= 255

def is_rgb_color(color):
    is_rgb = isinstance(color, list) or isinstance(color, tuple)
    is_rgb = is_rgb and len(color) == 3
    is_rgb = is_rgb and all([is_integer_color(el) for el in color])
    return is_rgb

def is_color(color):
    return is_string_color(color) or is_integer_color(color) or is_rgb_color(color)

def colorize(string, color = None, style = None, background = None, show = False): # it paints a text with given fullground and background color
    string = apply_ansi(string, background, 0)
    string = apply_ansi(string, color, 1)
    string = apply_ansi(string, style, 2)
    if show:
        print(string)
    return string

def uncolorize(string): # remove color codes from colored string
    colored = lambda: ansi_begin in string
    while colored():
        b = string.index(ansi_begin)
        e = string[b : ].index('m') + b + 1
        string = string.replace(string[b : e], '')
    return string

def apply_ansi(string, color, character):
    begin, end = ansi(color, character)
    return begin + string + end

#ansi_begin = '\033['
ansi_begin = '\x1b['
ansi_end = ansi_begin + '0m'

@memorize
def colors_to_ansi(fullground, style, background):
    color = [background, fullground, style]
    return ''.join([ansi(color[i], i)[0] for i in range(3)])

@memorize
def ansi(color, character):
    if color == no_color:
        return ['', '']
    col, fg, tp = '', '', ''
    if character == 2 and is_style(color):
        col = get_style_codes(color)
        col = ';'.join([str(el) for el in col])
    elif character != 2:
        fg = '38;' if character == 1 else '48;'
        tp = '5;'
        if is_string_color(color):
            col = str(get_color_code(color))
        elif is_integer_color(color):
            col = str(color)
        elif is_rgb_color(color):
            col = ';'.join([str(el) for el in color])
            tp = '2;'
    is_color = col != ''
    begin = ansi_begin + fg + tp + col + 'm' if is_color else ''
    end = ansi_end if is_color else ''
    return [begin, end]

## This section is useful to produce html colored version of the plot and to translate all color types (types 0 and 1) in rgb (type 2 in plotext) and avoid confusion. the match is almost exact and it depends on the terminal i suppose

def to_rgb(color):
    if is_string_color(color): # from 0 to 1
        color = get_color_code(color)
        #color = type0_to_type1_codes[code]
    if is_integer_color(color): # from 0 or 1 to 2
        return type1_to_type2_codes[color]
    return color

##############################################
############     Style Codes    ##############
##############################################

no_style = 'default'

styles = list(style_codes.keys()) + [no_style]

info_style = 'dim'

def get_style_code(style): # from single style to style number code
    style = style.strip()
    return style_codes[style]

def get_style_codes(style): # from many styles (separated by space) to as many number codes
    style = style.strip().split()
    codes = [get_style_code(el) for el in style if el in styles]
    codes = no_duplicates(codes)
    return codes

def get_style_name(code): # from style number code to style name
    codes = list(style_codes.values())
    return styles[codes.index(code)] if code in codes else no_style

def clean_styles(style): # it returns a well written sequence of styles (separated by spaces) from a possible confused one
    codes = get_style_codes(style)
    return ' '.join([get_style_name(el) for el in codes])

def is_style(style):
    style = style.strip().split() if isinstance(style, str) else ['']
    return any([el in styles for el in style])

##############################################
###########     Plot Utilities    ############
##############################################

def set_data(x = None, y = None): # it return properly formatted x and y data lists
   if x is None and y is None :
       x, y = [], []
   elif x is not None and y is None:
       y = x
       x = list(range(1, len(y) + 1))
   lx, ly = len(x), len(y)
   if lx != ly:
       l = min(lx, ly)
       x = x[ : l]
       y = y[ : l]
   return [list(x), list(y)]

##############################################
#######     Figure Class Utilities    ########
##############################################

def set_sizes(sizes, size_max): # given certain widths (or heights) - some of them are None -  it sets them so to respect max value
    bins = len(sizes)
    for s in range(bins):
        size_set = sum([el for el in sizes[0 : s] + sizes[s + 1 : ] if el is not None])
        available = max(size_max - size_set, 0)
        to_set = len([el for el in sizes[s : ] if el is None])
        sizes[s] = available // to_set if sizes[s] is None else sizes[s]
    return sizes

def fit_sizes(sizes, size_max): # honestly forgot the point of this function: yeeeeei :-) but it is useful - probably assumes all sizes not None (due to set_sizes) and reduces those that exceed size_max from last one to first
    bins = len(sizes)
    s = bins - 1
    #while (sum(sizes) != size_max if not_less else sum(sizes) > size_max) and s >= 0:
    while sum(sizes) > size_max and s >= 0:
        other_sizes = sum([sizes[i] for i in range(bins) if i != s])
        sizes[s] = max(size_max - other_sizes, 0)
        s -= 1
    return sizes

##############################################
#######     Build Class Utilities    #########
##############################################

def get_first(data, test = True): # if test take the first element, otherwise the second
    return data[0] if test else data[1]

def apply_scale(data, test = False): # apply log scale if test
    return log(data) if test else data

def reverse_scale(data, test = False): # apply log scale if test
    return power10(data) if test else data

def replace_none(data, num_data): # replace None elements in data with correspondent in num_data
    return [data[i] if data[i] is not None else num_data[i] for i in range(len(data))]

numerical = lambda el:  not (el is None or math.isnan(el)) or isinstance(el, str) # in the case of string datetimes
all_numerical = lambda data: all([numerical(el) for el in data])

def get_lim(data): # it returns the data minimum and maximum limits
    data = [el for el in data if numerical(el)]
    m = min(data, default = 0)
    M = max(data, default = 0)
    m, M = (m, M) if m != M else (0.5 * m, 1.5 * m) if m == M != 0 else (-1, 1)
    return [m, M]

def get_matrix_data(data, lim, bins): # from data to relative canvas coordinates
    change = lambda el: 0.5 + (bins - 1) * (el - lim[0]) / (lim[1] - lim[0])
    # round is so that for example 9.9999 = 10, otherwise the floor function will give different results
    return [math.floor(round(change(el), 8)) if numerical(el) else el for el in data]

def get_lines(x, y, *other): # it returns the lines between all couples of data points like x[i], y[i] to x[i + 1], y[i + 1]; other are the lisXt of markers and colors that needs to be elongated
    # if len(x) * len(y) == 0:
    #     return [], [], *[[]] * len(other)
    o = transpose(other, len(other))
    xl, yl, ol = [[] for i in range(3)]
    for n in range(len(x) - 1):
        xn, yn = x[n : n + 2], y[n : n + 2]
        xn, yn = get_line(xn, yn)
        xl += xn[:-1]
        yl += yn[:-1]
        ol += [o[n]] * len(xn[:-1])
    xl = xl + [x[-1]] if x != [] else xl
    yl = yl + [y[-1]] if x != [] else yl
    ol = ol + [o[-1]] if x != [] else ol
    return [xl, yl] + transpose(ol, len(other))

def get_line(x, y): # it returns a line of points from x[0],y[0] to x[1],y[1] distanced between each other in x and y by at least 1.
    if not all_numerical(join([x, y])):
        return x, y
    x0, x1 = x
    y0, y1 = y
    dx, dy = int(x1) - int(x0), int(y1) - int(y0)
    ax, ay = abs(dx), abs(dy)
    a = int(max(ax, ay) + 1)
    x = [int(el) for el in linspace(x0, x1, a)]
    y = [int(el) for el in linspace(y0, y1, a)]
    return [x, y]

def get_fill_level(fill, lim, bins):
    if fill is False:
        return False
    elif isinstance(fill, str):
        return fill
    else:
        fill = min(max(fill, lim[0]), lim[1])
        fill = get_matrix_data([fill], lim, bins)[0]
        return fill

def find_filling_values(x, y, y0):
    xn, yn, yf = [[]] * 3
    l = len(x);
    while len(x) > 0:
        i = len(xn)
        xn.append(x[i])
        yn.append(y[i])
        J = [j for j in range(l) if x[j] == x[i]]
        if J != []:
            Y = [y[j] for j in J]
            j = Y.index(min(Y))
            J.pop(j)
            [x.pop(j) for j in J]
            [y.pop(j) for j in J]
            yf.append(y[j])
    return xn, yn, yf

def get_fill_boundaries(x, y):
    xm = []
    l = len(x)
    for i in range(l):
        xi, yi = x[i], y[i]
        I = [j for j in range(l) if x[j] == xi and y[j] < yi]
        Y = [y[j] for j in I]
        m = min(Y, default = yi)
        xm.append([x[i], m])
    x, m = transpose(xm)
    return m

def fill_data(x, y, y0, *other): # it fills x, y with y data points reaching y0; and c are the list of markers and colors that needs to be elongated
    #y0 = get_fill_boundaries(x, y)
    y0 = get_fill_boundaries(x, y) if isinstance(y0, str) else [y0] * len(x)
    o = transpose(other, len(other))
    xf, yf, of = [[] for i in range(3)]
    xy = []
    for i in range(len(x)):
        xi, yi, y0i = x[i], y[i], y0[i]
        if [xi, yi] not in xy:
            xy.append([xi, yi])
            yn = range(y0i, yi + 1) if y0i < yi else range(yi, y0i) if y0i > yi else [y0i]
            yn = list(yn)
            xn = [xi] * len(yn)
            xf += xn
            yf += yn
            of += [o[i]] * len(xn)
    return [xf, yf] + transpose(of, len(other))

def remove_outsiders(x, y, width, height, *other):
    I = [i for i in range(len(x)) if x[i] in range(width) and y[i] in range(height)]
    o = transpose(other, len(other))
    return transpose([(x[i], y[i], *o[i]) for i in I], 2 + len(other))

def get_labels(ticks): # it returns the approximated string version of the data ticks
    d = distinguishing_digit(ticks)
    formatting_string = "{:." + str(d + 1) + "f}"
    labels = [formatting_string.format(el) for el in ticks]
    pos = [el.index('.') + d + 2 for el in labels]
    labels = [labels[i][: pos[i]] for i in range(len(labels))]
    all_integers = all(map(lambda el: el == int(el), ticks))
    labels = [add_extra_zeros(el, d) if len(labels) > 1 else el for el in labels] if not all_integers else [str(int(el)) for el in ticks]
    #sign = any([el < 0 for el in ticks])
    #labels = ['+' + labels[i] if ticks[i] > 0 and sign else labels[i] for i in range(len(labels))]
    return labels

def distinguishing_digit(data): # it return the minimum amount of decimal digits necessary to distinguish all elements of a list
    #data = [el for el in data if 'e' not in str(el)]
    d = [_distinguishing_digit(data[i], data[i + 1]) for i in range(len(data) - 1)]
    return max(d, default = 1)

def _distinguishing_digit(a, b): # it return the minimum amount of decimal digits necessary to distinguish a from b (when both are rounded to those digits).
    d = abs(a - b)
    d = 0 if d == 0 else - math.log10(2 * d)
    #d = round(d, 10)
    d = 0 if d < 0 else math.ceil(d)
    d = d + 1 if round(a, d) == round(b, d) else d
    return d

def add_extra_zeros(label, d): # it adds 0s at the end of a label if necessary
    zeros = len(label) - 1 - label.index('.' if 'e' not in label else 'e')
    if zeros < d:
        label += '0' * (d - zeros)
    return label

def add_extra_spaces(labels, side): # it adds empty spaces before or after the labels if necessary
    length = 0 if labels == [] else max_length(labels)
    if side == "left":
        labels = [space * (length - len(el)) + el for el in labels]
    if side == "right":
        labels = [el + space * (length - len(el)) for el in labels]
    return labels

def hd_group(x, y, xf, yf): # it returns the real coordinates of the HD markers and the matrix that defines the marker
    l, xfm, yfm = len(x), max(xf), max(yf)
    xm = [el // xfm if numerical(el) else el for el in x]
    ym = [el // yfm if numerical(el) else el for el in y]
    m = {}
    for i in range(l):
        xyi = xm[i], ym[i]
        xfi, yfi = xf[i], yf[i]
        mi = [[0 for x in range(xfi)] for y in range(yfi)]
        m[xyi] = mi
    for i in range(l):
        xyi = xm[i], ym[i]
        if all_numerical(xyi):
            xk, yk = x[i] % xfi, y[i] % yfi
            xk, yk = math.floor(xk), math.floor(yk)
            m[xyi][yk][xk] = 1
    x, y = transpose(m.keys(), 2)
    m = [tuple(join(el[::-1])) for el in m.values()]
    return x, y, m

###############################################
#############   Bar Functions    ##############
###############################################

def bars(x, y, width, minimum): # given the bars center coordinates and height, it returns the full bar coordinates
   # if x == []:
   #     return [], []
   bins = len(x)
   #bin_size_half = (max(x) - min(x)) / (bins - 1) * width / 2
   bin_size_half = width / 2
   # adjust the bar width according to the number of bins
   if bins > 1:
       bin_size_half *= (max(x) - min(x)) / (bins - 1)
   xbar, ybar = [], []
   for i in range(bins):
       xbar.append([x[i] - bin_size_half, x[i] + bin_size_half])
       ybar.append([minimum, y[i]])
   return xbar, ybar

def set_multiple_bar_data(*args):
    l = len(args)
    Y = [] if l == 0 else args[0] if l == 1 else args[1]
    Y = [Y] if not isinstance(Y, list) or len(Y) == 0 else Y
    m = len(Y[0])
    x = [] if l == 0 else list(range(1, m + 1)) if l == 1 else args[0]
    return x, Y

def hist_data(data, bins = 10, norm = False): # it returns data in histogram form if norm is False. Otherwise, it returns data in density form where all bins sum to 1.
    #data = [round(el, 15) for el in data]
    # if data == []:
    #     return [], []
    bins = 0 if len(data) == 0 else bins
    m, M = min(data, default = 0), max(data, default = 0)
    data = [(el - m) / (M - m) * bins if el != M else bins - 1 for el in data]
    data = [int(el) for el in data]
    histx = linspace(m, M, bins)
    histy = [0] * bins
    for el in data:
        histy[el] += 1
    if norm:
        histy = [el / len(data) for el in histy]
    return histx, histy

def single_bar(x, y, ylabel, marker, colors):
    l = len(y)
    lc = len(colors)
    xs = colorize(str(x), 'gray+', 'bold')
    bar = [marker * el for el in y]
    bar = [apply_ansi(bar[i], colors[i % lc], 1) for i in range(l)]
    ylabel = colorize(f'{int(round(ylabel))}', 'gray+', 'bold')
    bar = xs + space + ''.join(bar) + space + ylabel
    return bar

def bar_data(*args, width = None, mode = 'stacked'):
    x, Y = set_multiple_bar_data(*args)
    x = list(map(str, x))
    x = add_extra_spaces(x, 'right')
    lx = len(x[0])
    y = [sum(el) for el in transpose(Y)] if mode == 'stacked' else Y
    ly = max_length([round(el) for el in join(y)])

    width_term = terminal_width()
    width = width_term if width is None else min(width, width_term)
    width = max(width, lx + ly + 2 + 1)

    my = max(join(y))
    my = 1 if my == 0 else my
    dx = my / (width - lx - ly - 2)
    Yi = [[round(el / dx, 0) for el in y] for y in Y]
    Yi = transpose(Yi)

    return x, y, Yi, width

def correct_marker(marker = None):
    return simple_bar_marker if marker is None else marker[0]

def get_title(title, width):
    out = ''
    if title is not None:
        l = len(uncolorize(title))
        w1 = (width - 2 - l) // 2; w2 = width - l - 2 - w1
        l1 = '─' * w1 + space
        l2 = space + '─' * w2
        out = colorize(l1 + title + l2, 'gray+', 'bold') + '\n'
    return out

def get_simple_labels(marker, labels, colors, width):
    out = '\n'
    if labels != None:
        l = len(labels)
        lc = len(colors)
        out =  space.join([colorize(marker * 3, colors[i % lc]) + space + colorize(labels[i], 'gray+', 'bold') for i in range(l)])
        out = '\n' + get_title(out, width)
    return out

###############################################
#############   Box Functions    ##############
###############################################

def box(x, y, width, minimum): # given the bars center coordinates and height, it returns the full bar coordinates
    # if x == []:
    #     return [], []
    bins = len(x)
    #bin_size_half = (max(x) - min(x)) / (bins - 1) * width / 2
    bin_size_half = width / 2
    # adjust the bar width according to the number of bins
    if bins > 1:
        bin_size_half *= (max(x) - min(x)) / (bins - 1)
    c, q1, q2, q3, h, l = [], [], [], [], [], []
    xbar, ybar, mybar = [], [], []

    for i in range(bins):
        c.append(x[i])
        xbar.append([x[i] - bin_size_half, x[i] + bin_size_half])
        q1.append(quantile(y[i], 0.25))
        q2.append(quantile(y[i], 0.50))
        q3.append(quantile(y[i], 0.75))
        h.append(max(y[i]))
        l.append(min(y[i]))

    return q1, q2, q3, h, l, c, xbar

##############################################
##########    Image Utilities    #############
##############################################

def update_size(size_old, size_new): # it resize an image to the desired size, maintaining or not its size ratio and adding or not a pixel averaging factor with resample = True
    size_old = [size_old[0], size_old[1] / 2]
    ratio_old = size_old[1] / size_old[0]
    size_new = replace(size_new, size_old)
    ratio_new = size_new[1] / size_new[0]
    #ratio_new = size_new[1] / size_new[0]
    size_new = [1 if el == 0 else el for el in size_new]
    return [int(size_new[0]), int(size_new[1])]

def image_to_matrix(image): # from image to a matrix of pixels
    pixels = list(image.getdata())
    width, height = image.size
    return [pixels[i * width:(i + 1) * width] for i in range(height)]



from datetime import datetime as dt
from datetime import timezone as tz # useful for dates before 1970 in windows

class date_class():
    
    def __init__(self):
        self.date_form()
        self.time0 = 0
        self.set_time0('01/01/1900')

    def date_form(self, input_form = None, output_form = None): # it sets the datetime form used for functions that output date string and input date string
        input_form = 'd/m/Y' if input_form is None else input_form
        output_form = input_form if output_form is None else output_form
        self.input_form = self.correct_form(input_form)
        self.output_form = self.correct_form(output_form)
        
    def set_time0(self, string, input_form = None): # the origin of time, useful for log scale not to hit the 0 timestamp
        self.time0 = self.string_to_time(string, input_form, 0)

    def today_datetime(self): # today in datetime form
        return dt.today()
    
    def today_string(self, output_form = None): # today in string form
        return self.datetime_to_string(self.today_datetime(), output_form)
    
    def datetime_to_string(self, datetime, output_form = None): # from datetime form to string form
        output_form = self.output_form if output_form is None else self.correct_form(output_form)
        return datetime.strftime(output_form)
    
    def datetimes_to_strings(self, datetimes, output_form = None): # from datetime form to string form
        return [self.datetime_to_string(el, output_form) for el in datetimes]

    def string_to_datetime(self, string, input_form = None): # from date and times in string form to standard datetime input_form
        input_form = self.input_form if input_form is None else self.correct_form(input_form)
        return dt.strptime(string, input_form)

##############################################
############     Utilities    ################
##############################################

    def correct_form(self, date_form):
        return ''.join(['%' + el if el.isalpha() else el for el in date_form])

    def string_to_time(self, string, input_form = None, time0 = None):
        input_form = self.input_form if input_form is None else self.correct_form(input_form)
        time0 = self.time0 if time0 is None else time0
        try:
            return dt.strptime(string, input_form).replace(tzinfo = tz.utc).timestamp() - time0
        except:
            raise ValueError('Date Form should be: ' + input_form)

    def strings_to_time(self, strings, input_form = None):
        return [self.string_to_time(el, input_form) for el in strings]

    def time_to_string(self, time, output_form = None):
        return self.datetime_to_string(dt.fromtimestamp(time + self.time0), output_form)
    
    def times_to_string(self, times, input_form = None):
        return [self.time_to_string(el, input_form) for el in times]

